# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_gf_write(sisl_tmp, sisl_system):
    tb = sisl.Hamiltonian(sisl_system.gtb)
    f = sisl_tmp("file.TSGF")
    gf = sisl.io.get_sile(f)
    bz = sisl.MonkhorstPack(tb, [3, 3, 1])
    E = np.linspace(-2, 2, 20) + 1j * 1e-4
    S = np.eye(len(tb), dtype=np.complex128)

    gf.write_header(bz, E)
    for i, (ispin, new_hs, k, e) in enumerate(gf):
        Hk = tb.Hk(k, format="array")
        assert ispin == 0
        if new_hs and i % 2 == 0:
            gf.write_hamiltonian(Hk)
        elif new_hs:
            gf.write_hamiltonian(Hk, S)
        gf.write_self_energy(S * e - Hk)


def test_gf_write_read(sisl_tmp, sisl_system):
    tb = sisl.Hamiltonian(sisl_system.gtb)
    f = sisl_tmp("file.TSGF")

    bz = sisl.MonkhorstPack(tb, [3, 3, 1])
    E = np.linspace(-2, 2, 20) + 1j * 1e-4
    S = np.eye(len(tb), dtype=np.complex128)

    gf = sisl.io.get_sile(f)

    gf.write_header(bz, E)
    for i, (ispin, write_hs, k, e) in enumerate(gf):
        assert ispin == 0
        Hk = tb.Hk(k, format="array")
        if write_hs and i % 2 == 0:
            gf.write_hamiltonian(Hk)
        elif write_hs:
            gf.write_hamiltonian(Hk, S)
        gf.write_self_energy(S * e - Hk)

    nspin, no_u, k, E_file = gf.read_header()
    assert nspin == 1
    assert np.allclose(E, E_file)
    assert np.allclose(k, bz.k)

    for i, (ispin, write_hs, k, e) in enumerate(gf):
        assert ispin == 0
        Hk = tb.Hk(k, format="array")
        if write_hs and i % 2 == 0:
            Hk_file, _ = gf.read_hamiltonian()
        elif write_hs:
            Hk_file, Sk_file = gf.read_hamiltonian()
            assert np.allclose(S, Sk_file)
        assert np.allclose(Hk, Hk_file)

        SE_file = gf.read_self_energy()
        assert np.allclose(SE_file, S * e - Hk)


def test_gf_write_read_spin(sisl_tmp, sisl_system):
    f = sisl_tmp("file.TSGF")

    tb = sisl.Hamiltonian(sisl_system.gtb, spin=sisl.Spin("P"))
    tb.construct([(0.1, 1.5), ([0.1, -0.1], [2.7, 1.6])])

    bz = sisl.MonkhorstPack(tb, [3, 3, 1])
    E = np.linspace(-2, 2, 3) + 1j * 1e-4
    S = np.eye(len(tb), dtype=np.complex128)

    gf = sisl.io.get_sile(f)

    gf.write_header(bz, E)
    for i, (ispin, write_hs, k, e) in enumerate(gf):
        Hk = tb.Hk(k, spin=ispin, format="array")
        if write_hs and i % 2 == 0:
            gf.write_hamiltonian(Hk)
        elif write_hs:
            gf.write_hamiltonian(Hk, S)
        gf.write_self_energy(S * e - Hk)

    # Check it isn't opened
    assert not gf._fortran_is_open()

    nspin, no_u, k, E_file = gf.read_header()
    assert nspin == 2
    assert np.allclose(E, E_file)
    assert np.allclose(k, bz.k)

    for i, (ispin, write_hs, k, e) in enumerate(gf):
        Hk = tb.Hk(k, spin=ispin, format="array")
        if write_hs and i % 2 == 0:
            Hk_file, _ = gf.read_hamiltonian()
        elif write_hs:
            Hk_file, Sk_file = gf.read_hamiltonian()
            assert np.allclose(S, Sk_file)
        assert np.allclose(Hk, Hk_file)

        SE_file = gf.read_self_energy()
        assert np.allclose(SE_file, S * e - Hk)


def test_gf_write_read_direct(sisl_tmp, sisl_system):
    f = sisl_tmp("file.TSGF")

    tb = sisl.Hamiltonian(sisl_system.gtb, spin=sisl.Spin("P"))
    tb.construct([(0.1, 1.5), ([0.1, -0.1], [2.7, 1.6])])

    bz = sisl.MonkhorstPack(tb, [3, 3, 1])
    E = np.linspace(-2, 2, 3) + 1j * 1e-4
    S = np.eye(len(tb), dtype=np.complex128)

    gf = sisl.io.get_sile(f)

    gf.write_header(bz, E)
    for i, (ispin, write_hs, k, e) in enumerate(gf):
        Hk = tb.Hk(k, spin=ispin, format="array")
        if write_hs and i % 2 == 0:
            gf.write_hamiltonian(Hk)
        elif write_hs:
            gf.write_hamiltonian(Hk, S)
        gf.write_self_energy(S * e - Hk)

    # ensure it is not opened
    assert not gf._fortran_is_open()

    # First try from beginning
    for e in [0, 1, E[1], 0, E[0]]:
        ie = gf.Eindex(e)
        SE1 = gf.self_energy(e, bz.k[2, :])
        assert gf._state == 1
        assert gf._ik == 2
        assert gf._iE == ie
        assert gf._ispin == 0
        assert gf._is_read == 1

        SE2 = gf.self_energy(e, bz.k[2, :], spin=1)
        assert gf._state == 1
        assert gf._ik == 2
        assert gf._iE == ie
        assert gf._ispin == 1
        assert gf._is_read == 1

        assert not np.allclose(SE1, SE2)

        # In the middle we read some hamiltonians
        H1, S1 = gf.HkSk(bz.k[2, :], spin=0)
        assert gf._state == 0
        assert gf._ik == 2
        assert gf._iE == 0
        assert gf._ispin == 0
        assert gf._is_read == 1
        assert np.allclose(S, S1)

        H2, S1 = gf.HkSk(bz.k[2, :], spin=1)
        assert gf._state == 0
        assert gf._ik == 2
        assert gf._iE == 0
        assert gf._ispin == 1
        assert gf._is_read == 1
        assert np.allclose(S, S1)
        assert not np.allclose(H1, H2)
        assert not np.allclose(H1, SE1)

        H2, S1 = gf.HkSk(bz.k[2, :], spin=0)
        assert gf._state == 0
        assert gf._ik == 2
        assert gf._iE == 0
        assert gf._ispin == 0
        assert gf._is_read == 1
        assert np.allclose(S, S1)
        assert np.allclose(H1, H2)

        # Now read self-energy
        SE2 = gf.self_energy(e, bz.k[2, :], spin=0)
        assert gf._state == 1
        assert gf._ik == 2
        assert gf._iE == ie
        assert gf._ispin == 0
        assert gf._is_read == 1

        assert np.allclose(SE1, SE2)


def test_gf_sile_error():
    with pytest.raises(sisl.SileError):
        sisl.get_sile("non_existing_file.TSGF").read_header()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.utils._arrays import *

pytestmark = pytest.mark.utils


def test_batched_indices_2d():
    ref = np.random.rand(10, 3)
    y = ref[[0, 3, 5]]
    assert np.allclose(batched_indices(ref, y)[0], [0, 3, 5])
    assert np.allclose(batched_indices(ref, y[1])[0], [3])


def test_batched_indices_3d():
    ref = np.random.rand(10, 3, 3)
    y = np.random.rand(3, 3)
    y = ref[[0, 3, 5], [0, 1, 2]]
    idx = batched_indices(ref, y)
    assert np.allclose(idx[0], [0, 3, 5])
    assert np.allclose(idx[1], [0, 1, 2])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_si_pdos_kgrid_tsde_dm(sisl_files):
    fdf = sisl.get_sile(
        sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"),
        base=sisl_files("siesta", "Si_pdos_k"),
    )

    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSDE"))

    DM1 = si.read_density_matrix(geometry=fdf.read_geometry())
    DM2 = fdf.read_density_matrix(order=["TSDE"])

    Ef1 = si.read_fermi_level()
    Ef2 = fdf.read_fermi_level()

    assert Ef1 == Ef2

    assert DM1._csr.spsame(DM2._csr)
    assert np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])


def test_si_pdos_kgrid_tsde_edm(sisl_files):
    fdf = sisl.get_sile(
        sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"),
        base=sisl_files("siesta", "Si_pdos_k"),
    )
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSDE"))

    EDM1 = si.read_energy_density_matrix(geometry=fdf.read_geometry())
    EDM2 = fdf.read_energy_density_matrix(order=["TSDE"])

    assert EDM1._csr.spsame(EDM2._csr)
    assert np.allclose(EDM1._csr._D[:, :-1], EDM2._csr._D[:, :-1])


@pytest.mark.filterwarnings("ignore", message="*Casting complex values")
@pytest.mark.parametrize(("matrix"), ["density", "energy_density"])
def test_si_pdos_kgrid_tsde_edm_dtypes(sisl_files, sisl_tmp, matrix):
    fdf = sisl.get_sile(
        sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"),
        base=sisl_files("siesta", "Si_pdos_k"),
    )
    data = []
    mull = None

    for dtype in (np.float32, np.float64, np.complex64, np.complex128):
        M = getattr(fdf, f"read_{matrix}_matrix")(dtype=dtype)
        data.append(M)
        assert M.dtype == dtype

        if mull is None:
            mull = M.mulliken()
        else:
            assert np.allclose(mull, M.mulliken(), atol=1e-5)

    fnc = sisl_tmp("tmp.nc")
    for M in data:
        M.write(fnc)
        # The overlap should be here...
        M1 = M.read(fnc)
        assert np.allclose(mull, M1.mulliken(), atol=1e-5)


@pytest.mark.filterwarnings("ignore", message="*wrong sparse pattern")
def test_si_pdos_kgrid_tsde_dm_edm_rw(sisl_files, sisl_tmp):
    fdf = sisl.get_sile(
        sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"),
        base=sisl_files("siesta", "Si_pdos_k"),
    )
    geom = fdf.read_geometry()
    f1 = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSDE"))

    DM1 = f1.read_density_matrix(geometry=geom)
    EDM1 = f1.read_energy_density_matrix(geometry=geom)

    f2 = sisl.get_sile(sisl_tmp("noEf.TSDE"))
    # by default everything gets sorted...
    f2.write_density_matrices(DM1, EDM1, sort=False)
    DM2 = f2.read_density_matrix(geometry=geom)
    EDM2 = f2.read_energy_density_matrix(geometry=geom)
    assert DM1._csr.spsame(DM2._csr)
    assert np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])
    assert EDM1._csr.spsame(EDM2._csr)
    assert np.allclose(EDM1._csr._D[:, :-1], EDM2._csr._D[:, :-1])

    # Now the matrices ARE finalized, we don't have to do anything again
    EDM2 = EDM1.copy()
    EDM2.shift(-2.0, DM1)
    f3 = sisl.get_sile(sisl_tmp("Ef.TSDE"))
    f3.write_density_matrices(DM1, EDM2, Ef=-2.0, sort=False)
    DM3 = f3.read_density_matrix(geometry=geom)
    EDM3 = f3.read_energy_density_matrix(geometry=geom)
    assert DM1._csr.spsame(DM3._csr)
    assert np.allclose(DM1._csr._D[:, :-1], DM3._csr._D[:, :-1])
    assert EDM1._csr.spsame(EDM3._csr)
    assert np.allclose(EDM1._csr._D[:, :-1], EDM3._csr._D[:, :-1])

    f3.write_density_matrices(DM1, EDM2, Ef=-2.0, sort=True)
    DM3 = f3.read_density_matrix(geometry=geom, sort=False)
    EDM3 = f3.read_energy_density_matrix(geometry=geom, sort=False)
    assert DM1._csr.spsame(DM3._csr)
    assert not np.allclose(DM1._csr._D[:, :-1], DM3._csr._D[:, :-1])
    DM3.finalize()
    assert np.allclose(DM1._csr._D[:, :-1], DM3._csr._D[:, :-1])
    assert EDM1._csr.spsame(EDM3._csr)
    assert not np.allclose(EDM1._csr._D[:, :-1], EDM3._csr._D[:, :-1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl
from sisl.viz.processors.eigenstate import (
    create_wf_grid,
    eigenstate_geometry,
    get_eigenstate,
    get_grid_nsc,
    project_wavefunction,
    tile_if_k,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(scope="module", params=["Gamma", "X"])
def k(request):
    if request.param == "Gamma":
        return (0, 0, 0)
    elif request.param == "X":
        return (0.5, 0, 0)


@pytest.fixture(scope="module")
def graphene():
    r = np.linspace(0, 3.5, 50)
    f = np.exp(-r)

    orb = sisl.AtomicOrbital("2pzZ", (r, f))
    return sisl.geom.graphene(orthogonal=True, atoms=sisl.Atom(6, orb))


@pytest.fixture(scope="module")
def eigenstate(k, graphene):
    # Create a simple graphene tight binding Hamiltonian
    H = sisl.Hamiltonian(graphene)
    H.construct([(0.1, 1.44), (0, -2.7)])

    return H.eigenstate(k=k)


def test_get_eigenstate(eigenstate, graphene):
    sel_eigenstate = get_eigenstate(eigenstate, 2)

    assert sel_eigenstate.state.shape == (1, graphene.no)
    assert np.allclose(sel_eigenstate.state, eigenstate.state[2])

    eigenstate = eigenstate.copy()

    eigenstate.info["index"] = np.array([0, 3, 1, 2])

    sel_eigenstate = get_eigenstate(eigenstate, 2)

    assert sel_eigenstate.state.shape == (1, graphene.no)
    assert np.allclose(sel_eigenstate.state, eigenstate.state[3])


def test_eigenstate_geometry(eigenstate, graphene):
    # It should give us the geometry associated with the eigenstate
    assert eigenstate_geometry(eigenstate) is graphene

    # Unless we provide a geometry
    graphene_copy = graphene.copy()
    assert eigenstate_geometry(eigenstate, graphene_copy) is graphene_copy


def test_tile_if_k(eigenstate, graphene):
    # If the eigenstate is calculated at gamma, we don't need to tile
    tiled_geometry = tile_if_k(graphene, (2, 2, 2), eigenstate)

    if eigenstate.info["k"] == (0, 0, 0):
        # If the eigenstate is calculated at gamma, we don't need to tile
        assert tiled_geometry is graphene
    elif eigenstate.info["k"] == (0.5, 0, 0):
        # If the eigenstate is calculated at X, we need to tile
        # but only the first lattice vector.
        assert tiled_geometry is not graphene
        assert np.allclose(tiled_geometry.cell, graphene.cell * (2, 1, 1))


def test_get_grid_nsc(eigenstate):
    grid_nsc = get_grid_nsc((2, 2, 2), eigenstate)

    if eigenstate.info["k"] == (0, 0, 0):
        assert grid_nsc == (2, 2, 2)
    elif eigenstate.info["k"] == (0.5, 0, 0):
        assert grid_nsc == (1, 2, 2)


def test_create_wf_grid(eigenstate, graphene):
    new_graphene = graphene.copy()
    grid = create_wf_grid(eigenstate, grid_prec=0.2, geometry=new_graphene)

    assert isinstance(grid, sisl.Grid)
    assert grid.geometry is new_graphene

    # Check that the datatype is correct
    if eigenstate.info["k"] == (0, 0, 0):
        assert grid.grid.dtype == np.float64
    else:
        assert grid.grid.dtype == np.complex128

    # Check that the grid precision is right.
    assert np.allclose(np.linalg.norm(grid.dcell, axis=1), 0.2, atol=0.01)

    provided_grid = sisl.Grid(0.2, geometry=new_graphene, dtype=np.float64)

    grid = create_wf_grid(eigenstate, grid=provided_grid)

    assert grid is provided_grid


def test_project_wavefunction(eigenstate, graphene):
    k = eigenstate.info["k"]

    grid = project_wavefunction(eigenstate[2], geometry=graphene)

    assert isinstance(grid, sisl.Grid)

    # Check that the datatype is correct
    if k == (0, 0, 0):
        assert grid.grid.dtype == np.float64
    else:
        assert grid.grid.dtype == np.complex128

    # Check that the grid is not empty
    assert not np.allclose(grid.grid, 0)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import scipy.linalg as sl

from sisl.linalg import inv, inv_destroy

pytestmark = [pytest.mark.linalg, pytest.mark.inv]


def test_inv1():
    np.random.seed(1204982)
    a = np.random.rand(10, 10)
    ac = a.copy()
    xs = sl.inv(a)
    x = inv(a)
    assert np.allclose(xs, x)
    assert np.allclose(a, ac)


def test_inv_d1():
    np.random.seed(1204982)
    a = np.random.rand(10, 10)
    xs = sl.inv(a)
    x = inv_destroy(a)
    assert np.allclose(xs, x)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

import sisl
from sisl import Lattice
from sisl.viz.processors.coords import (
    coords_depth,
    project_to_axes,
    projected_1D_data,
    projected_1Dcoords,
    projected_2D_data,
    projected_2Dcoords,
    projected_3D_data,
    sphere,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(scope="module", params=["numpy", "lattice"])
def Cell(request):
    if request.param == "numpy":
        return np.array
    elif request.param == "lattice":
        return Lattice


@pytest.fixture(scope="module")
def coords_dataset():
    geometry = sisl.geom.bcc(2.93, "Au", False)

    return xr.Dataset(
        {"xyz": (("atom", "axis"), geometry.xyz)},
        coords={"axis": [0, 1, 2]},
        attrs={"geometry": geometry},
    )


def test_projected_1D_coords(Cell):
    cell = Cell([[0, 1, 0], [1, 0, 0], [0, 0, 1]])

    x, y, z = 3, -4, 2

    coords = np.array([[x, y, z]])

    # Project to cartesian
    projected = projected_1Dcoords(cell, coords, "x")
    assert np.allclose(projected, [[x]])

    projected = projected_1Dcoords(cell, coords, "-y")
    assert np.allclose(projected, [[-y]])

    # Project to lattice
    projected = projected_1Dcoords(cell, coords, "b")
    assert np.allclose(projected, [[x]])

    # Project to vector
    projected = projected_1Dcoords(cell, coords, [x, 0, z])
    assert np.allclose(projected, [[1]])


def test_projected_2D_coords(Cell):
    cell = Cell([[0, 1, 0], [1, 0, 0], [0, 0, 1]])

    x, y, z = 3, -4, 2

    coords = np.array([[x, y, z]])

    # Project to cartesian
    projected = projected_2Dcoords(cell, coords, "x", "y")
    assert np.allclose(projected, [[x, y]])

    projected = projected_2Dcoords(cell, coords, "-x", "y")
    assert np.allclose(projected, [[-x, y]])

    projected = projected_2Dcoords(cell, coords, "z", "x")
    assert np.allclose(projected, [[z, x]])

    # Project to lattice
    projected = projected_2Dcoords(cell, coords, "a", "b")
    assert np.allclose(projected, [[y, x]])

    projected = projected_2Dcoords(cell, coords, "-b", "a")
    assert np.allclose(projected, [[-x, y]])

    # Project to vectors
    projected = projected_2Dcoords(cell, coords, [x, y, 0], [0, 0, z])
    assert np.allclose(projected, [[1, 1]])


def test_coords_depth(coords_dataset):
    depth = coords_depth(coords_dataset, ["x", "y"])
    assert isinstance(depth, np.ndarray)
    assert np.allclose(depth, coords_dataset.xyz.sel(axis=2).values)

    depth = coords_depth(coords_dataset, ["y", "x"])
    assert np.allclose(depth, -coords_dataset.xyz.sel(axis=2).values)

    depth = coords_depth(coords_dataset, [[1, 0, 0], [0, 0, 1]])
    assert np.allclose(depth, -coords_dataset.xyz.sel(axis=1).values)


@pytest.mark.parametrize("center", [[0, 0, 0], [1, 1, 0]])
def test_sphere(center):
    coords = sphere(center=center, r=3.5, vertices=15)

    assert isinstance(coords, dict)

    assert "x" in coords
    assert "y" in coords
    assert "z" in coords

    assert coords["x"].shape == coords["y"].shape == coords["z"].shape == (15**2,)

    R = np.linalg.norm(
        np.array([coords["x"], coords["y"], coords["z"]]).T - center, axis=1
    )

    assert np.allclose(R, 3.5)


def test_projected_1D_data(coords_dataset):
    # No data
    projected = projected_1D_data(coords_dataset, "y")
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, coords_dataset.xyz.sel(axis=1))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, 0)

    # Data from function
    projected = projected_1D_data(coords_dataset, "-y", dataaxis_1d=np.sin)
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, -coords_dataset.xyz.sel(axis=1))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, np.sin(-coords_dataset.xyz.sel(axis=1)))

    # Data from array
    projected = projected_1D_data(
        coords_dataset, "-y", dataaxis_1d=coords_dataset.xyz.sel(axis=2).values
    )
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, -coords_dataset.xyz.sel(axis=1))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, coords_dataset.xyz.sel(axis=2))


def test_projected_2D_data(coords_dataset):
    projected = projected_2D_data(coords_dataset, "-y", "x")
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, -coords_dataset.xyz.sel(axis=1))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, coords_dataset.xyz.sel(axis=0))

    assert "depth" in projected.data_vars

    projected = projected_2D_data(coords_dataset, "x", "y", sort_by_depth=True)
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, coords_dataset.xyz.sel(axis=0))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, coords_dataset.xyz.sel(axis=1))

    assert "depth" in projected.data_vars
    assert np.allclose(projected.depth.values, coords_dataset.xyz.sel(axis=2))
    # Check that points are sorted by depth.
    assert np.all(np.diff(projected.depth) > 0)


def test_projected_3D_data(coords_dataset):
    projected = projected_3D_data(coords_dataset)
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, coords_dataset.xyz.sel(axis=0))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, coords_dataset.xyz.sel(axis=1))
    assert "z" in projected.data_vars
    assert np.allclose(projected.z, coords_dataset.xyz.sel(axis=2))


def test_project_to_axes(coords_dataset):
    projected = project_to_axes(coords_dataset, ["z"], dataaxis_1d=4)
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, coords_dataset.xyz.sel(axis=2))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, 4)
    assert "z" not in projected.data_vars

    projected = project_to_axes(coords_dataset, ["-y", "x"])
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, -coords_dataset.xyz.sel(axis=1))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, coords_dataset.xyz.sel(axis=0))
    assert "z" not in projected.data_vars

    projected = project_to_axes(coords_dataset, ["x", "y", "z"])
    assert isinstance(projected, xr.Dataset)
    assert "x" in projected.data_vars
    assert np.allclose(projected.x, coords_dataset.xyz.sel(axis=0))
    assert "y" in projected.data_vars
    assert np.allclose(projected.y, coords_dataset.xyz.sel(axis=1))
    assert "z" in projected.data_vars
    assert np.allclose(projected.z, coords_dataset.xyz.sel(axis=2))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry, get_sile
from sisl.io.siesta.struct import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_struct1(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.STRUCT_IN")
    sisl_system.g.write(structSileSiesta(f, "w"))
    g = structSileSiesta(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)
    assert sisl_system.g.atoms.equal(g.atoms, R=False)


def test_struct_reorder(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.STRUCT_IN")
    g = sisl_system.g.copy()
    g.atoms[0] = Atom(1)
    g.write(structSileSiesta(f, "w"))
    g2 = structSileSiesta(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)


def test_struct_ghost(sisl_tmp):
    f = sisl_tmp("ghost.STRUCT_IN")
    a1 = Atom(1)
    am1 = Atom(-1)
    g = Geometry([[0.0, 0.0, i] for i in range(2)], [a1, am1], 2.0)
    g.write(structSileSiesta(f, "w"))

    g2 = structSileSiesta(f).read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert np.allclose(g.atoms.Z, g2.atoms.Z)
    assert g.atoms[0].__class__ is g2.atoms[0].__class__
    assert g.atoms[1].__class__ is g2.atoms[1].__class__
    assert g.atoms[0].__class__ is not g2.atoms[1].__class__


def test_si_pdos_kgrid_struct_out(sisl_files):
    fdf = get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    struct = get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.STRUCT_OUT"))

    struct_geom = struct.read_geometry()
    fdf_geom = fdf.read_geometry(order="STRUCT")

    assert np.allclose(struct_geom.cell, fdf_geom.cell)
    assert np.allclose(struct_geom.xyz, fdf_geom.xyz)

    struct_sc = struct.read_lattice()
    fdf_sc = fdf.read_lattice(order="STRUCT")

    assert np.allclose(struct_sc.cell, fdf_sc.cell)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from io import StringIO

import pytest

from sisl.io.siesta.basis import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_si_ion_nc(sisl_files):
    pytest.importorskip("netCDF4")
    f = sisl_files("siesta", "Si_pdos_k", "Si.ion.nc")
    with ionncSileSiesta(f) as sile:
        atom = sile.read_basis()

    # Number of orbitals
    assert len(atom) == 9


def test_si_ion_xml(sisl_files):
    f = sisl_files("siesta", "Si_pdos_k", "Si.ion.xml")
    with ionxmlSileSiesta(f) as sile:
        atom = sile.read_basis()

    # Number of orbitals
    assert len(atom) == 9


def test_si_ion_xml_handle(sisl_files):
    f = open(sisl_files("siesta", "Si_pdos_k", "Si.ion.xml"), "r")

    with ionxmlSileSiesta(f) as sile:
        assert "Buffer" in sile.__class__.__name__
        atom = sile.read_basis()

    # Number of orbitals
    assert len(atom) == 9


def test_si_ion_xml_stringio(sisl_files):
    f = StringIO(open(sisl_files("siesta", "Si_pdos_k", "Si.ion.xml"), "r").read())

    with ionxmlSileSiesta(f) as sile:
        assert "Buffer" in sile.__class__.__name__
        atom = sile.read_basis()

    # Number of orbitals
    assert len(atom) == 9


def test_si_ion_compare(sisl_files):
    pytest.importorskip("netCDF4")
    f = sisl_files("siesta", "Si_pdos_k", "Si.ion.nc")
    with ionncSileSiesta(f) as sile:
        nc = sile.read_basis()

    f = sisl_files("siesta", "Si_pdos_k", "Si.ion.xml")
    with ionxmlSileSiesta(f) as sile:
        xml = sile.read_basis()

    assert nc == xml



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import scipy.linalg as sl

from sisl.linalg import invsqrth, lowdin, signsqrt, sqrth

pytestmark = [pytest.mark.linalg]


def test_sqrth():
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    a = a + a.T
    np.fill_diagonal(a, 4)
    sa1 = sl.sqrtm(a)
    sa2 = sqrth(a)
    assert np.allclose(sa1, sa2)
    assert np.allclose(sa2 @ sa2, a)


def test_sqrth_negative():
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    a = a + a.T
    np.fill_diagonal(a, a.diagonal() - 1)
    sa1 = sl.sqrtm(a)
    sa2 = sqrth(a)
    assert np.allclose(sa1, sa2)
    assert np.allclose(sa2 @ sa2, a)


def test_invsqrth():
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    a = a + a.T
    np.fill_diagonal(a, 4)
    sa1 = sl.inv(sl.sqrtm(a))
    sa2 = invsqrth(a)
    assert np.allclose(sa1, sa2)


def test_invsqrth_offset():
    # offsetting eigenvalues only works if the matrix is
    # positive semi-definite
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    a = a + a.T
    # without diagonal filling
    np.fill_diagonal(a, a.diagonal() + 4)
    sa1 = invsqrth(a)

    # similar thing, with diagonal
    np.fill_diagonal(a, a.diagonal() + 4)
    eig, ev = sl.eigh(a)
    eig = signsqrt(eig - 4)
    np.divide(1, eig, where=(eig != 0), out=eig)
    sa2 = (ev * eig) @ ev.conj().T
    assert np.allclose(sa1, sa2)


@pytest.mark.parametrize("driver", ["eigh", "gesdd", "gesvd", "schur"])
def test_lowdin(driver):
    # offsetting eigenvalues only works if the matrix is
    # positive semi-definite
    np.random.seed(1285947159)

    b = np.random.rand(10, 10)
    b = b + b.T
    np.fill_diagonal(b, b.diagonal() + 4)

    # Recreate an overlap matrix with all eigenvalues being 1.
    _, ev = sl.eigh(b)
    b = ev.T @ ev

    # Create "H"
    a = np.random.rand(10, 10)
    a = a + a.T

    eig, ev = sl.eigh(a, b)

    # Calculate eigenvalues for a, b and then
    # by the Lowdin transformed a
    aL = lowdin(b, a, driver=driver)
    eigL, evL = sl.eigh(aL)

    assert np.allclose(eig, eigL)
    assert not np.allclose(ev, evL)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """

import warnings

import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.tbtrans]

netCDF4 = pytest.importorskip("netCDF4")


@pytest.mark.slow
@pytest.mark.filterwarnings("ignore", message="*.*.o2p")
def test_1_graphene_all_content(sisl_files):
    """This tests manifolds itself as:

    sisl.geom.graphene(orthogonal=False).tile(2, 0).tile(3, 0)

    All output is enabled:

    ### FDF ###
    # Transmission related quantities
    TBT.T.All T
    TBT.T.Out T
    TBT.T.Eig 2
    TBT.T.Orbital T

    # Density of states
    TBT.DOS.Elecs T
    TBT.DOS.Gf T
    TBT.DOS.A T
    TBT.DOS.A.All T

    # Orbital transmissions and Crystal-Orbital investigations.
    TBT.Symmetry.TimeReversal F
    TBT.Current.Orb T
    TBT.COOP.Gf T
    TBT.COOP.A T
    TBT.COHP.Gf T
    TBT.COHP.A T

    TBT.k [1 11 1]
    ### FDF ###
    """
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    assert tbt.E.min() > -1.0
    assert tbt.E.max() < 1.0
    # We have 400 energy-points
    ne = len(tbt.E)
    assert ne == 4
    assert tbt.ne == ne

    # We have 100 k-points
    nk = len(tbt.kpt)
    assert nk == 11
    assert tbt.nk == nk
    assert tbt.wk.sum() == pytest.approx(1.0)

    for i in range(ne):
        assert tbt.Eindex(tbt.E[i]) == i

    # Check raises
    with pytest.warns(sisl.SislWarning):
        tbt.Eindex(tbt.E.min() - 1.0)
    with pytest.warns(sisl.SislInfo):
        tbt.Eindex(tbt.E.min() - 2e-3)
    with pytest.warns(sisl.SislWarning):
        tbt.kindex([0, 0, 0.5])

    for i in range(nk):
        assert tbt.kindex(i) == i
        assert tbt.kindex(tbt.kpt[i]) == i

    # Get geometry
    geom = tbt.geometry
    geom_c1 = tbt.read_geometry(atoms=sisl.Atoms(sisl.Atom[6], geom.na))
    geom_c2 = tbt.read_geometry(atoms=sisl.Atoms(sisl.Atom(6, orbs=2), geom.na))
    assert geom_c1 == geom_c2

    norbs = 4

    # Check read is the same as the direct query
    assert tbt.na == geom.na
    assert tbt.no == geom.no
    assert tbt.no != geom.na
    assert tbt.na == 2 * 3 * 4
    assert np.allclose(tbt.cell, geom.cell)

    # Check device atoms (1-orbital system)
    assert tbt.na_d != tbt.no_d
    assert tbt.na_d * norbs == tbt.no_d
    assert tbt.na_d == 2 * 3 * 2  # device has not two ends
    assert len(tbt.pivot()) == 2 * 3 * 2 * norbs
    assert len(tbt.pivot(in_device=True)) == len(tbt.pivot())
    assert np.all(tbt.pivot(in_device=True, sort=True) == np.arange(tbt.no_d))
    assert np.all(tbt.pivot(sort=True) == np.sort(tbt.pivot()))

    # Just check they are there
    assert tbt.n_btd() == len(tbt.btd())

    # Check electrodes
    assert len(tbt.elecs) == 2
    elecs = tbt.elecs[:]
    assert elecs == ["Left", "Right"]
    for i, elec in enumerate(elecs):
        assert tbt._elec(i) == elec

    # Check the chemical potentials
    assert tbt.eta() == pytest.approx(0, abs=1e-7)
    for elec in elecs:
        assert tbt.n_btd(elec) == len(tbt.btd(elec))
        assert tbt.chemical_potential(elec) == pytest.approx(0.0)
        assert tbt.electron_temperature(elec) == pytest.approx(300.0, abs=1)
        assert tbt.eta(elec) == pytest.approx(1e-6, abs=1e-8)

    # Check electrode relevant stuff
    left = elecs[0]
    right = elecs[1]

    # Assert we have transmission symmetry
    assert np.allclose(tbt.transmission(left, right), tbt.transmission(right, left))
    assert np.allclose(
        tbt.transmission_eig(left, right), tbt.transmission_eig(right, left)
    )
    # Check that the total transmission is larger than the sum of transmission eigenvalues
    assert np.all(
        tbt.transmission(left, right) + 1e-7
        >= tbt.transmission_eig(left, right).sum(-1)
    )
    assert np.all(
        tbt.transmission(right, left) + 1e-7
        >= tbt.transmission_eig(right, left).sum(-1)
    )

    # Check that we can't retrieve from same to same electrode
    with pytest.raises(ValueError):
        tbt.transmission(left, left)
    with pytest.raises(ValueError):
        tbt.transmission_eig(left, left)

    assert np.allclose(
        tbt.transmission(left, right, kavg=False),
        tbt.transmission(right, left, kavg=False),
    )

    # Both methods should be identical for simple bulk systems
    # However, when dealing with DFT and a relatively large eta,
    # they tend to be too different
    assert np.allclose(
        tbt.reflection(left), tbt.reflection(left, from_single=True), atol=1e-5
    )

    # Also check for each k
    for ik in range(nk):
        assert np.allclose(
            tbt.transmission(left, right, ik), tbt.transmission(right, left, ik)
        )
        assert np.allclose(
            tbt.transmission_eig(left, right, ik), tbt.transmission_eig(right, left, ik)
        )
        assert np.all(
            tbt.transmission(left, right, ik) + 1e-7
            >= tbt.transmission_eig(left, right, ik).sum(-1)
        )
        assert np.all(
            tbt.transmission(right, left, ik) + 1e-7
            >= tbt.transmission_eig(right, left, ik).sum(-1)
        )
        assert np.allclose(
            tbt.DOS(kavg=ik),
            tbt.ADOS(left, kavg=ik) + tbt.ADOS(right, kavg=ik),
            atol=1e-5,
        )
        assert np.allclose(
            tbt.DOS(E=0.195, kavg=ik),
            tbt.ADOS(left, E=0.195, kavg=ik) + tbt.ADOS(right, E=0.195, kavg=ik),
            atol=1e-5,
        )

    # Check that norm returns correct values
    assert tbt.norm() == 1
    assert tbt.norm(norm="all") == tbt.no_d
    assert tbt.norm(norm="atom") == tbt.norm(norm="orbital")

    # Check atom is equivalent to orbital
    for norm in ("atom", "orbital"):
        assert tbt.norm(0, norm=norm) == 0.0
        assert tbt.norm(3 * 2, norm=norm) == norbs
        assert tbt.norm(range(3 * 2, 3 * 3), norm=norm) == 3 * norbs

    # Assert sum(ADOS) == DOS
    assert np.allclose(tbt.DOS(), tbt.ADOS(left) + tbt.ADOS(right), atol=1e-5)
    assert np.allclose(
        tbt.DOS(sum=False),
        tbt.ADOS(left, sum=False) + tbt.ADOS(right, sum=False),
        atol=1e-5,
    )

    # Now check orbital resolved DOS
    assert np.allclose(
        tbt.DOS(sum=False),
        tbt.ADOS(left, sum=False) + tbt.ADOS(right, sum=False),
        atol=1e-5,
    )

    # Current must be 0 when the chemical potentials are equal
    assert tbt.current(left, right) == pytest.approx(0.0)
    assert tbt.current(right, left) == pytest.approx(0.0)

    high_low = tbt.current_parameter(left, 0.5, 0.0025, right, -0.5, 0.0025)
    low_high = tbt.current_parameter(left, -0.5, 0.0025, right, 0.5, 0.0025)
    assert high_low > 0.0
    assert low_high < 0.0
    assert -high_low == pytest.approx(low_high)
    with pytest.warns(sisl.SislWarning):
        tbt.current_parameter(left, -10.0, 0.0025, right, 10.0, 0.0025)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        # Since this is a perfect system there should be *no* QM shot-noise
        # Also, the shot-noise is related to the applied bias, so NO shot-noise
        assert np.allclose(
            (tbt.shot_noise(left, right, kavg=False) * tbt.wkpt.reshape(-1, 1)).sum(0),
            0.0,
        )
        assert np.allclose(tbt.shot_noise(left, right), 0.0)
        assert np.allclose(tbt.shot_noise(right, left), 0.0)
        assert np.allclose(tbt.shot_noise(left, right, kavg=1), 0.0)

        # Since the data-file does not contain all T-eigs (only the first two)
        # we can't correctly calculate the fano factors
        # This is a pristine system, hence all fano-factors should be more or less zero
        # All transmissions are step-functions, however close to band-edges there is some
        # smearing.
        # When calculating the Fano factor it is extremely important that the zero_T is *sufficient*
        # I don't now which value is *good*
        assert np.all(
            (tbt.fano(left, right, kavg=False) * tbt.wkpt.reshape(-1, 1)).sum(0) <= 1
        )
        assert np.all(tbt.fano(left, right) <= 1)
        assert np.all(tbt.fano(right, left) <= 1)
        assert np.all(tbt.fano(left, right, kavg=0) <= 1)

        # Neither should the noise_power exist
        assert (
            tbt.noise_power(right, left, kavg=False) * tbt.wkpt
        ).sum() == pytest.approx(0.0)
        assert tbt.noise_power(right, left) == pytest.approx(0.0)
        assert tbt.noise_power(right, left, kavg=0) == pytest.approx(0.0)

    # Check specific DOS queries
    DOS = tbt.DOS
    ADOS = tbt.ADOS

    assert DOS(2.0, atoms=True, sum=False).size == geom.names["Device"].size
    assert np.allclose(
        DOS(2.0, atoms="Device", sum=False), DOS(2.0, atoms=True, sum=False)
    )
    assert DOS(2.0, orbitals=True, sum=False).size == geom.a2o("Device", all=True).size
    assert ADOS(left, 2.0, atoms=True, sum=False).size == geom.names["Device"].size
    assert (
        ADOS(left, 2.0, orbitals=True, sum=False).size
        == geom.a2o("Device", all=True).size
    )
    assert np.allclose(
        ADOS(left, 2.0, atoms="Device", sum=False),
        ADOS(left, 2.0, atoms=True, sum=False),
    )

    for E in (None, 2.0, 3.0):
        assert np.allclose(DOS(E), ADOS(left, E) + ADOS(right, E), atol=1e-5)

        for o, indices in (("atoms", range(8, 12)), ("orbitals", range(50, 70))):
            opt = {o: indices}

            for sum in (False, True):
                opt["sum"] = sum
                assert np.allclose(
                    DOS(E, **opt),
                    ADOS(left, E, **opt) + ADOS(right, E, **opt),
                    atol=1e-5,
                )

            opt["norm"] = o[:-1]
            assert np.allclose(
                DOS(E, **opt), ADOS(left, E, **opt) + ADOS(right, E, **opt), atol=1e-5
            )

    # Check orbital currents
    E = 0.75
    # Sum of orbital current should be 0 (in == out)
    # However, when eta is *big*, this tends to get blurred.
    orb_left = tbt.orbital_transmission(E, left)
    orb_right = tbt.orbital_transmission(E, right)
    assert orb_left.sum() == pytest.approx(0.0, abs=1e-5)
    assert orb_right.sum() == pytest.approx(0.0, abs=1e-5)

    d1 = geom.a2o(range(6, 12), all=True).reshape(-1, 1)
    d2 = geom.a2o(range(12, 18), all=True).reshape(-1, 1)
    # TODO, this does not hold here... why?
    # assert orb_left[d1, d2.T].sum() == pytest.approx(tbt.transmission(left, right)[E])
    # assert orb_left[d1, d2.T].sum() == pytest.approx(-orb_left[d2, d1.T].sum())
    # assert orb_right[d2, d1.T].sum() == pytest.approx(tbt.transmission(right, left)[E])
    # assert orb_right[d2, d1.T].sum() == pytest.approx(-orb_right[d1, d2.T].sum())

    orb_left.sort_indices()
    atom_left = tbt.bond_transmission(E, left, what="all")
    atom_left.sort_indices()
    assert np.allclose(atom_left.data, tbt.sparse_orbital_to_atom(orb_left).data)
    orb_right.sort_indices()
    atom_right = tbt.bond_transmission(E, right, what="all")
    atom_right.sort_indices()
    assert np.allclose(atom_right.data, tbt.sparse_orbital_to_atom(orb_right).data)

    # Calculate the atom current
    # For 1-orbital systems the activity and non-activity are equivalent
    assert np.allclose(
        tbt.atom_transmission(E, left), tbt.atom_transmission(E, left, activity=False)
    )
    tbt.vector_transmission(E, left)
    assert np.allclose(
        tbt.sparse_atom_to_vector(atom_left) / 2,
        tbt.vector_transmission(E, left, what="all"),
    )

    # Check COOP curves
    coop = tbt.orbital_COOP(E)
    coop_l = tbt.orbital_ACOOP(E, left)
    coop_r = tbt.orbital_ACOOP(E, right)
    coop_lr = coop_l + coop_r

    # Ensure aligment
    coop.eliminate_zeros()
    coop.sorted_indices()
    coop_lr.eliminate_zeros()
    coop_lr.sorted_indices()
    assert np.allclose(coop.data, coop_lr.data)

    coop = tbt.orbital_COOP(E, isc=[0, 0, 0])
    coop_l = tbt.orbital_ACOOP(E, left, isc=[0, 0, 0])
    coop_r = tbt.orbital_ACOOP(E, right, isc=[0, 0, 0])
    coop_lr = coop_l + coop_r

    coop.eliminate_zeros()
    coop.sorted_indices()
    coop_lr.eliminate_zeros()
    coop_lr.sorted_indices()
    assert np.allclose(coop.data, coop_lr.data)

    coop = tbt.atom_COOP(E)
    coop_l = tbt.atom_ACOOP(E, left)
    coop_r = tbt.atom_ACOOP(E, right)
    coop_lr = coop_l + coop_r

    coop.eliminate_zeros()
    coop.sorted_indices()
    coop_lr.eliminate_zeros()
    coop_lr.sorted_indices()
    assert np.allclose(coop.data, coop_lr.data)

    coop = tbt.atom_COOP(E, isc=[0, 0, 0])
    coop_l = tbt.atom_ACOOP(E, left, isc=[0, 0, 0])
    coop_r = tbt.atom_ACOOP(E, right, isc=[0, 0, 0])
    coop_lr = coop_l + coop_r

    coop.eliminate_zeros()
    coop.sorted_indices()
    coop_lr.eliminate_zeros()
    coop_lr.sorted_indices()
    assert np.allclose(coop.data, coop_lr.data)

    # Check COHP curves
    coop = tbt.orbital_COHP(E)
    coop_l = tbt.orbital_ACOHP(E, left)
    coop_r = tbt.orbital_ACOHP(E, right)
    coop_lr = coop_l + coop_r

    coop.eliminate_zeros()
    coop.sorted_indices()
    coop_lr.eliminate_zeros()
    coop_lr.sorted_indices()
    assert np.allclose(coop.data, coop_lr.data)

    coop = tbt.atom_COHP(E)
    coop_l = tbt.atom_ACOHP(E, left)
    coop_r = tbt.atom_ACOHP(E, right)
    coop_lr = coop_l + coop_r

    coop.eliminate_zeros()
    coop.sorted_indices()
    coop_lr.eliminate_zeros()
    coop_lr.sorted_indices()
    assert np.allclose(coop.data, coop_lr.data)

    # Simply print out information
    tbt.info()
    for elec in elecs:
        tbt.info(elec)


@pytest.mark.filterwarnings("ignore", message="closest energy")
def test_1_graphene_Eindex(sisl_files):
    # gh-829 (Eindex -> only energy resolutions)
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    idx = tbt.Eindex(-0.7)
    assert idx == 0
    assert tbt.E[idx] == pytest.approx(-0.75)
    idx = tbt.Eindex(-0.7, method="below")
    assert idx == 0
    assert tbt.E[idx] == pytest.approx(-0.75)
    idx = tbt.Eindex(-0.7, method="above")
    assert idx == 1
    assert tbt.E[idx] == pytest.approx(-0.25)

    with pytest.raises(ValueError):
        tbt.Eindex(-0.76, method="below")
    with pytest.raises(ValueError):
        tbt.Eindex(0.76, method="above")


@pytest.mark.slow
def test_1_graphene_all_tbtav(sisl_files, sisl_tmp):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    f = sisl_tmp("graphene.TBT.AV.nc")
    tbt.write_tbtav(f)


def test_1_graphene_all_fail_kavg(sisl_files, sisl_tmp):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    with pytest.raises(ValueError):
        tbt.transmission(kavg=[0, 1])


def test_1_graphene_sparse_current(sisl_files, sisl_tmp):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))

    J = tbt.orbital_current()
    assert np.allclose(J.data, 0)

    # convert to bond_current
    j = tbt.sparse_orbital_to_atom(J)
    assert np.allclose(j.data, 0)

    j = tbt.sparse_atom_to_vector(j)
    assert np.allclose(j.data, 0)

    j = tbt.sparse_orbital_to_scalar(J, activity=True)
    assert np.allclose(j.data, 0)

    J = tbt.orbital_current(what="+")
    assert np.allclose(J.data, 0)


@pytest.mark.filterwarnings("ignore:.*requesting energy")
def test_1_graphene_all_fail_kavg_E(sisl_files, sisl_tmp):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    with pytest.raises(ValueError):
        tbt.orbital_COOP(kavg=[0, 1], E=0.1)


def test_1_graphene_all_ArgumentParser(sisl_files, sisl_tmp):
    pytest.importorskip("matplotlib", reason="matplotlib not available")
    import matplotlib as mpl

    mpl.rcParams["text.usetex"] = False

    # Local routine to run the collected actions
    def run(ns):
        ns._actions_run = True
        # Run all so-far collected actions
        for A, Aargs, Akwargs in ns._actions:
            A(*Aargs, **Akwargs)
        ns._actions_run = False
        ns._actions = []

    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    geom = tbt.read_geometry()

    p, ns = tbt.ArgumentParser()
    p.parse_args([], namespace=ns)

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--energy", " -0.75:0.75"], namespace=ns)
    assert not out._actions_run
    run(out)

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--kpoint", "1"], namespace=ns)
    assert out._krng
    run(out)
    assert out._krng == 1

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--norm", "orbital"], namespace=ns)
    run(out)
    assert out._norm == "orbital"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--norm", "atom"], namespace=ns)
    run(out)
    assert out._norm == "atom"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--kpoint", "1", "--norm", "orbital"], namespace=ns)
    run(out)
    assert out._krng == 1
    assert out._norm == "orbital"

    #
    # Atoms 7-12 are in the device region
    #

    p, ns = tbt.ArgumentParser()
    atom = "10:11,24"
    out = p.parse_args(["--atom", atom], namespace=ns)
    run(out)
    assert out._Ovalue == atom
    assert np.all(out._Orng == geom.a2o([9, 10], all=True))

    p, ns = tbt.ArgumentParser()
    atom = "10:11,12,24:30"
    out = p.parse_args(["--atom", atom], namespace=ns)
    run(out)
    assert out._Ovalue == atom
    # Only 13-48 is in the device
    assert np.all(out._Orng == geom.a2o([9, 10, 11], all=True))

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--transmission", "Left", "Right"], namespace=ns)
    run(out)
    assert len(out._data) == 2
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][0] == "T"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(
        ["--transmission", "Left", "Right", "--transmission-bulk", "Left"], namespace=ns
    )
    run(out)
    assert len(out._data) == 3
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][0] == "T"
    assert out._data_header[2][:2] == "BT"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--dos", "--dos", "Left", "--ados", "Right"], namespace=ns)
    run(out)
    assert len(out._data) == 4
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][0] == "D"
    assert out._data_header[2][:2] == "AD"
    assert out._data_header[3][:2] == "AD"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--bulk-dos", "Left", "--ados", "Right"], namespace=ns)
    run(out)
    assert len(out._data) == 3
    assert out._data_header[0][0] == "E"
    assert out._data_header[1][:2] == "BD"
    assert out._data_header[2][:2] == "AD"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--transmission-eig", "Left", "Right"], namespace=ns)
    run(out)
    assert out._data_header[0][0] == "E"
    for i in range(1, len(out._data)):
        assert out._data_header[i][:4] == "Teig"

    p, ns = tbt.ArgumentParser()
    out = p.parse_args(["--info"], namespace=ns)

    # Test output
    f = sisl_tmp("graphene.dat")
    p, ns = tbt.ArgumentParser()
    out = p.parse_args(
        ["--transmission-eig", "Left", "Right", "--out", f], namespace=ns
    )
    assert len(out._data) == 0

    f1 = sisl_tmp("graphene_1.dat")
    f2 = sisl_tmp("graphene_2.dat")
    p, ns = tbt.ArgumentParser()
    out = p.parse_args(
        [
            "--transmission",
            "Left",
            "Right",
            "--out",
            f1,
            "--dos",
            "--atom",
            "8:2:48",
            "--dos",
            "Right",
            "--ados",
            "Left",
            "--out",
            f2,
        ],
        namespace=ns,
    )

    d = sisl.io.tableSile(f1).read_data()
    assert len(d) == 2
    d = sisl.io.tableSile(f2).read_data()
    assert len(d) == 4
    assert np.allclose(d[1, :], (d[2, :] + d[3, :]) * 2, atol=1e-3)
    assert np.allclose(d[2, :], d[3, :], atol=1e-3)

    f = sisl_tmp("graphene_T.png")
    p, ns = tbt.ArgumentParser()
    out = p.parse_args(
        ["--transmission", "Left", "Right", "--transmission-bulk", "Left", "--plot", f],
        namespace=ns,
    )


# Requesting an orbital outside of the device region
def test_1_graphene_all_warn_orbital(sisl_files):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    with pytest.warns(sisl.SislWarning):
        tbt.o2p(1)


# Requesting an atom outside of the device region
def test_1_graphene_all_warn_atom(sisl_files):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))
    with pytest.warns(sisl.SislWarning):
        tbt.a2p(1)


def test_1_graphene_all_sparse_data_isc_request(sisl_files):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))

    # get supercell with isc
    sc = tbt.read_lattice()

    # request the full matrix
    for elec in [0, 1]:
        J_all = tbt.orbital_transmission(0.75, elec)
        J_all.eliminate_zeros()

        # Ensure we actually have something
        assert J_all.nnz > 0

        # partial summed isc
        # Test that the full matrix and individual access is the same
        J_sum = sum(tbt.orbital_transmission(0.75, elec, isc=isc) for isc in sc.sc_off)
        assert J_sum.nnz == J_all.nnz
        assert (J_sum - J_all).nnz == 0


def test_1_graphene_all_sparse_data_orbitals(sisl_files):
    tbt = sisl.get_sile(sisl_files("siesta", "tbtrans", "graphene", "graphene.TBT.nc"))

    # request the full matrix
    J_all = tbt.orbital_transmission(0.75, 0)
    J_12 = tbt.orbital_transmission(0.75, 0, orbitals=[2, 3])

    assert J_12.nnz < J_all.nnz // 2


def test_projection_raise(sisl_files):
    tbt = sisl.get_sile(
        sisl_files("siesta", "tbtrans", "c60_projection", "projection.TBT.nc")
    )

    with pytest.raises(sisl.io.tbtrans.MissingFDFTBtransError):
        tbt.orbital_transmission(1.995, 0)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" Global sisl fixtures """
import logging
import os
from functools import partial
from pathlib import Path

import numpy as np
import pytest

from sisl import Atom, Geometry, Hamiltonian, Lattice, _environ
from sisl._help import has_module

# Here we create the necessary methods and fixtures to enabled/disable
# tests depending on whether a sisl-files directory is present.

_log = logging.getLogger(__name__)

sisl_files_tests = _environ.get_environ_variable("SISL_FILES_TESTS")
if (sisl_files_tests / "tests").is_dir():
    sisl_files_tests = sisl_files_tests / "tests"


# Modify items based on whether the env is correct or not
def pytest_collection_modifyitems(config, items):
    global sisl_files_tests
    if sisl_files_tests.is_dir():
        return

    xfail_sisl_files = pytest.mark.xfail(
        run=False,
        reason="requires env(SISL_FILES_TESTS) pointing to clone of: https://github.com/zerothi/sisl-files",
    )
    for item in items:
        # Only skip those that have the sisl_files fixture
        # GLOBAL skipping of ALL tests that don't have this fixture
        if "sisl_files" in item.fixturenames:
            item.add_marker(xfail_sisl_files)


@pytest.fixture(scope="function")
def sisl_tmp(request, tmp_path_factory):
    """sisl specific temporary file and directory creator.

        sisl_tmp(file, dir_name='sisl')
        sisl_tmp.file(file, dir_name='sisl')
        sisl_tmp.dir('sisl')

    The scope of the `sisl_tmp` fixture is at a function level to
    clean up after each function.
    """
    global __file__
    # The test file is in a tests/test_<name>.py,
    # so step back twice
    path = Path(request.node.path).parent.parent
    # This file is in sisl/conftest.py
    # And we wish to retain the sisl path, so go two levels up as well
    path = str(path.relative_to(Path(__file__).parent.parent))

    # Now path should be something like:
    # sisl/io/siesta

    class FileFactory:
        def __init__(self):
            self.base = tmp_path_factory.getbasetemp()
            self.dirs = [self.base]
            self.files = []

        def dir(self, name=path):
            # Make name a path
            D = Path(name.replace(os.path.sep, "-"))
            if not (self.base / D).is_dir():
                # tmp_path_factory.mktemp returns pathlib.Path
                self.dirs.append(tmp_path_factory.mktemp(str(D), numbered=False))

            return self.dirs[-1]

        def file(self, name, dir_name=path):
            # self.base *is* a pathlib
            D = self.base / dir_name.replace(os.path.sep, "-")
            if D in self.dirs:
                i = self.dirs.index(D)
            else:
                self.dir(dir_name)
                i = -1
            self.files.append(self.dirs[i] / name)
            return str(self.files[-1])

        def getbase(self):
            return self.dirs[-1]

        def __call__(self, name, dir_name=path):
            """Shorthand for self.file"""
            return self.file(name, dir_name)

        def teardown(self):
            while len(self.files) > 0:
                # Do each removal separately
                f = self.files.pop()
                if f.is_file():
                    try:
                        f.close()
                    except Exception:
                        pass
                    try:
                        f.unlink()
                    except Exception:
                        pass
            while len(self.dirs) > 0:
                # Do each removal separately (from back of directory)
                d = self.dirs.pop()
                if d.is_dir():
                    try:
                        d.rmdir()
                    except Exception:
                        pass

    ff = FileFactory()
    request.addfinalizer(ff.teardown)
    return ff


@pytest.fixture(scope="session")
def sisl_files():
    """Environment catcher for the large files hosted in a different repository.

    If SISL_FILES_TESTS has been defined in the environment variable the directory
    will be used for the tests with this as a fixture.

    If the environment variable is empty and a test has this fixture, it will
    be skipped.
    """
    global sisl_files_tests

    if sisl_files_tests.is_dir():

        def _path(*files):
            p = sisl_files_tests.joinpath(*files)
            if p.exists():
                return p
            _log.info("sisl_files: test requested non-existing ' {p!s}' -> xfail")

            # I expect this test to fail due to the wrong environment.
            # But it isn't an actual fail since it hasn't runned...
            pytest.xfail(
                reason=f"Environment SISL_FILES_TESTS may point to a wrong path(?); file {p} not found",
            )

    else:
        _log.info(
            "sisl_files SISL_FILES_TESTS={sisl_files_tests!s} does not exist, xfail dependencies"
        )

        def _path(*files):
            pytest.xfail(
                reason=f"Environment SISL_FILES_TESTS not pointing to a valid directory.",
            )

    return _path


@pytest.fixture(scope="session")
def sisl_tolerance():
    r32 = (1e-6, 1e-11)
    r64 = (1e-9, 1e-15)
    return {
        np.float32: r32,
        np.float32(0).dtype: r32,
        np.float64: r64,
        np.float64(0).dtype: r64,
        np.complex64: r32,
        np.complex64(0).dtype: r32,
        np.complex128: r64,
        np.complex128(0).dtype: r64,
        None: r64,
    }


@pytest.fixture(scope="session")
def sisl_allclose(sisl_tolerance):
    def factory(dtype):
        atol, rtol = sisl_tolerance[dtype]
        return partial(np.allclose, atol=atol, rtol=rtol)

    return {key: factory(key) for key in sisl_tolerance.keys()}


@pytest.fixture(scope="session")
def sisl_isclose(sisl_tolerance):
    def factory(dtype):
        atol, rtol = sisl_tolerance[dtype]
        return partial(np.isclose, atol=atol, rtol=rtol)

    return {key: factory(key) for key in sisl_tolerance.keys()}


@pytest.fixture(scope="session", params=[np.float32, np.float64])
def sisl_float(request, sisl_tolerance):
    yield (request.param,) + sisl_tolerance[request.param]


@pytest.fixture(scope="session", params=[np.complex64, np.complex128])
def sisl_complex(request, sisl_tolerance):
    yield (request.param,) + sisl_tolerance[request.param]


@pytest.fixture(scope="session")
def sisl_system():
    """A preset list of geometries/Hamiltonians."""

    class System:
        pass

    d = System()

    alat = 1.42
    sq3h = 3.0**0.5 * 0.5
    C = Atom(Z=6, R=1.42)
    lattice = Lattice(
        np.array([[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64)
        * alat,
        nsc=[3, 3, 1],
    )
    d.g = Geometry(
        np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * alat,
        atoms=C,
        lattice=lattice,
    )

    d.R = np.array([0.1, 1.5])
    d.t = np.array([0.0, 2.7])
    d.tS = np.array([(0.0, 1.0), (2.7, 0.0)])
    d.C = Atom(Z=6, R=max(d.R))
    d.lattice = Lattice(
        np.array([[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64)
        * alat,
        nsc=[3, 3, 1],
    )
    d.gtb = Geometry(
        np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * alat,
        atoms=C,
        lattice=lattice,
    )

    d.ham = Hamiltonian(d.gtb)
    d.ham.construct([(0.1, 1.5), (0.1, 2.7)])
    return d


collect_ignore = []
collect_ignore_glob = []

# We are ignoring stuff in sisl.viz if nodify cannot be imported
# skip paths
_skip_paths = []
if not has_module("nodify"):
    _skip_paths.append(os.path.join("sisl", "viz"))


def pytest_ignore_collect(collection_path, config):
    # ensure we only compare against final *sisl* stuff
    global _skip_paths
    parts = list(Path(collection_path).parts)
    parts.reverse()
    sisl_parts = parts[: parts.index("sisl")]
    sisl_parts.reverse()
    sisl_path = str(Path("sisl").joinpath(*sisl_parts))

    for skip_path in _skip_paths:
        if skip_path in sisl_path:
            return True
    return False


def pytest_report_header(config, start_path):
    global sisl_files_tests
    global _skip_paths
    s = []
    s.append(f"sisl-test: found FILES_TESTS: {sisl_files_tests!s}")
    if _skip_paths:
        s.append("sisl-test: skip test-discovery in these folders:")
        for skip in _skip_paths:
            s.append(f" - {skip}")
    return s


def pytest_configure(config):
    pytest.sisl_travis_skip = pytest.mark.skipif(
        os.environ.get("SISL_TRAVIS_CI", "false").lower() == "true",
        reason="running on TRAVIS",
    )

    # Locally manage pytest.ini input
    for mark in [
        "io",
        "generic",
        "bloch",
        "hamiltonian",
        "geometry",
        "geom",
        "neighbor",
        "shape",
        "state",
        "electron",
        "phonon",
        "utils",
        "unit",
        "distribution",
        "spin",
        "self_energy",
        "help",
        "messages",
        "namedindex",
        "sparse",
        "lattice",
        "supercell",
        "sc",
        "quaternion",
        "sparse_geometry",
        "sparse_orbital",
        "ranges",
        "physics",
        "physics_feature",
        "orbital",
        "oplist",
        "grid",
        "atoms",
        "atom",
        "periodictable",
        "sgrid",
        "sdata",
        "sgeom",
        "version",
        "bz",
        "brillouinzone",
        "monkhorstpack",
        "bandstructure",
        "inv",
        "eig",
        "linalg",
        "densitymatrix",
        "dynamicalmatrix",
        "energydensitymatrix",
        "siesta",
        "tbtrans",
        "dftb",
        "vasp",
        "w90",
        "wannier90",
        "gulp",
        "fdf",
        "fhiaims",
        "aims",
        "orca",
        "collection",
        "category",
        "geom_category",
        "plot",
        "slow",
        "overlap",
        "mixing",
        "typing",
        "only",
        "viz",
        "processors",
        "data",
        "plots",
        "plotters",
    ]:
        config.addinivalue_line(
            "markers", f"{mark}: mark test to run only on named environment"
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl.io.siesta import aniSileSiesta

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_ani(sisl_tmp):
    f = sisl_tmp("sisl.ANI")
    open(f, "w").write(
        """1

C   0.00000000  0.00000000  0.00000000
2

C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
3

C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
C   2.00000  0.00000000  0.00000000
4

C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
C   2.00000  0.00000000  0.00000000
C   3.00000  0.00000000  0.00000000
"""
    )
    a = aniSileSiesta(f)
    g = a.read_geometry[:]()
    assert len(g) == 4
    assert g[0].na == 1

    g = a.read_geometry[1:]()
    assert len(g) == 3
    assert g[0].na == 2

    g = a.read_geometry[:]()
    assert len(g) == 4
    assert g[0].na == 1 and g[2].na == 3

    g = a.read_geometry[1::1]()
    assert len(g) == 3
    assert g[0].na == 2 and g[1].na == 3

    g = a.read_geometry[::2]()
    assert len(g) == 2
    assert g[0].na == 1 and g[1].na == 3

    g = a.read_geometry[:2:1]()
    assert len(g) == 2
    assert g[0].na == 1 and g[1].na == 2

    g = a.read_geometry[1:]()
    assert g[0].na == 2 and len(g) == 3

    g = a.read_geometry[1]()
    assert g.na == 2

    g = a.read_geometry[1:3:1]()
    assert g[1].na == 3 and len(g) == 2

    g = a.read_geometry[1:3:2]()
    assert g[0].na == 2 and len(g) == 1

    g = a.read_geometry[-2:]()
    assert g[0].na == 3 and len(g) == 2
    assert g[1].na == 4

    g = a.read_geometry[-2::-1]()
    assert g[0].na == 3 and len(g) == 3
    assert g[1].na == 2
    assert g[2].na == 1

    g = a.read_geometry(lattice=None, atoms=None)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from sisl import Lattice
from sisl.viz.processors.cell import (
    cell_to_lines,
    gen_cell_dataset,
    infer_cell_axes,
    is_1D_cartesian,
    is_cartesian_unordered,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(scope="module", params=["numpy", "lattice"])
def Cell(request):
    if request.param == "numpy":
        return np.array
    elif request.param == "lattice":
        return Lattice


def test_cartesian_unordered(Cell):
    assert is_cartesian_unordered(
        Cell([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
    )

    assert is_cartesian_unordered(
        Cell([[0, 0, 1], [1, 0, 0], [0, 1, 0]]),
    )

    assert not is_cartesian_unordered(
        Cell([[0, 2, 1], [1, 0, 0], [0, 1, 0]]),
    )


def test_1D_cartesian(Cell):
    assert is_1D_cartesian(Cell([[1, 0, 0], [0, 1, 0], [0, 0, 1]]), "x")

    assert is_1D_cartesian(Cell([[1, 0, 0], [0, 1, 1], [0, 0, 1]]), "x")

    assert not is_1D_cartesian(Cell([[1, 0, 0], [1, 1, 0], [0, 0, 1]]), "x")


def test_infer_cell_axes(Cell):
    assert infer_cell_axes(
        Cell([[0, 1, 0], [1, 0, 0], [0, 0, 1]]), axes=["x", "y", "z"]
    ) == [1, 0, 2]

    assert infer_cell_axes(
        Cell([[0, 1, 0], [1, 0, 0], [0, 0, 1]]), axes=["b", "y"]
    ) == [1, 0]


def test_gen_cell_dataset():
    lattice = Lattice([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    cell_dataset = gen_cell_dataset(lattice)

    assert isinstance(cell_dataset, xr.Dataset)

    assert "lattice" in cell_dataset.attrs
    assert cell_dataset.attrs["lattice"] is lattice

    assert "xyz" in cell_dataset.data_vars
    assert cell_dataset.xyz.shape == (2, 2, 2, 3)
    assert np.all(cell_dataset.xyz.values == lattice.vertices())


@pytest.mark.parametrize("mode", ["box", "axes", "other"])
def test_cell_to_lines(mode):
    lattice = Lattice([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    cell_dataset = gen_cell_dataset(lattice)

    if mode == "other":
        with pytest.raises(ValueError):
            cell_to_lines(cell_dataset, mode)
    else:
        lines = cell_to_lines(cell_dataset, mode, cell_style={"color": "red"})

        assert isinstance(lines, xr.Dataset)

        if mode == "box":
            # 19 points are required to draw a box
            assert lines.xyz.shape == (19, 3)
        elif mode == "axes":
            # 9 points are required to draw the axes
            assert lines.xyz.shape == (9, 3)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest

import sisl
from sisl import geom
from sisl.io import SileError, fdfSileSiesta
from sisl.messages import SislWarning
from sisl.unit.siesta import unit_convert

pytestmark = [
    pytest.mark.io,
    pytest.mark.siesta,
    pytest.mark.fdf,
    pytest.mark.filterwarnings("ignore", message="*number of supercells"),
]


def test_fdf1(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.fdf")
    sisl_system.g.write(fdfSileSiesta(f, "w"))

    fdf = fdfSileSiesta(f)
    str(fdf)
    with fdf:
        fdf.readline()

        # Be sure that we can read it in a loop
        assert fdf.get("LatticeConstant") > 0.0
        assert fdf.get("LatticeConstant") > 0.0
        assert fdf.get("LatticeConstant") > 0.0

        fdf.read_lattice()
        fdf.read_geometry()


def test_fdf2(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.fdf")
    sisl_system.g.write(fdfSileSiesta(f, "w"))
    g = fdfSileSiesta(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)
    for ia in g:
        assert g.atoms[ia].Z == sisl_system.g.atoms[ia].Z
        assert g.atoms[ia].tag == sisl_system.g.atoms[ia].tag


def test_fdf_units(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.fdf")
    fdf = fdfSileSiesta(f, "w")
    g = sisl_system.g

    for unit in ["bohr", "ang", "fractional", "frac"]:
        fdf.write_geometry(g, unit=unit)
        g2 = fdfSileSiesta(f).read_geometry()
        assert np.allclose(g.cell, g2.cell)
        assert np.allclose(g.xyz, g2.xyz)
        for ia in g:
            assert g.atoms[ia].Z == g2.atoms[ia].Z
            assert g.atoms[ia].tag == g2.atoms[ia].tag


def test_lattice(sisl_tmp):
    f = sisl_tmp("file.fdf")
    lines = [
        "Latticeconstant 1. Ang",
        "%block Latticevectors",
        " 1. 1. 1.",
        " 0. 0. 1.",
        " 1. 0. 1.",
        "%endblock",
    ]
    with open(f, "w") as fh:
        fh.write("\n".join(lines))

    cell = np.array([[1.0] * 3, [0, 0, 1], [1, 0, 1]])
    lattice = fdfSileSiesta(f).read_lattice()
    assert np.allclose(lattice.cell, cell)

    lines = [
        "Latticeconstant 1. Bohr",
        "%block Latticevectors",
        " 1. 1. 1.",
        " 0. 0. 1.",
        " 1. 0. 1.",
        "%endblock",
    ]
    with open(f, "w") as fh:
        fh.write("\n".join(lines))

    lattice = fdfSileSiesta(f).read_lattice()
    assert np.allclose(lattice.cell, cell * unit_convert("Bohr", "Ang"))

    cell = np.diag([2.0] * 3)
    lines = [
        "Latticeconstant 2. Ang",
        "%block Latticeparameters",
        " 1. 1. 1. 90. 90. 90.",
        "%endblock",
    ]
    with open(f, "w") as fh:
        fh.write("\n".join(lines))

    lattice = fdfSileSiesta(f).read_lattice()
    assert np.allclose(lattice.cell, cell)


def test_lattice_fail(sisl_tmp):
    f = sisl_tmp("file.fdf")
    lines = [
        "%block Latticevectors",
        " 1. 1. 1.",
        " 0. 0. 1.",
        " 1. 0. 1.",
        "%endblock",
    ]
    with open(f, "w") as fh:
        fh.write("\n".join(lines))
    with pytest.raises(SileError):
        fdfSileSiesta(f).read_lattice()


def test_geometry(sisl_tmp):
    f = sisl_tmp("file.fdf")
    sc_lines = [
        "Latticeconstant 1. Ang",
        "%block latticeparameters",
        " 1. 1. 1. 90. 90. 90.",
        "%endblock",
    ]
    lines = [
        "NumberOfAtoms 2",
        "%block chemicalSpeciesLabel",
        " 1 6 C",
        " 2 12 H",
        "%endblock",
        "AtomicCoordinatesFormat Ang",
        "%block atomiccoordinatesandatomicspecies",
        " 1. 1. 1. 1",
        " 0. 0. 1. 1",
        " 1. 0. 1. 2",
        "%endblock",
    ]

    with open(f, "w") as fh:
        fh.write("\n".join(sc_lines) + "\n")
        fh.write("\n".join(lines))

    fdf = fdfSileSiesta(f, base=sisl_tmp.getbase())
    g = fdf.read_geometry()
    assert g.na == 2
    assert np.allclose(g.xyz, [[1.0] * 3, [0, 0, 1]])
    assert np.allclose(g.atoms.Z, [6, 6])
    assert g.atoms.nspecies == 2

    # default read # of atoms from list
    with open(f, "w") as fh:
        fh.write("\n".join(sc_lines) + "\n")
        fh.write("\n".join(lines[1:]))

    fdf = fdfSileSiesta(f, base=sisl_tmp.getbase())
    g = fdf.read_geometry()
    assert g.na == 3
    assert np.allclose(g.xyz, [[1.0] * 3, [0, 0, 1], [1, 0, 1]])
    assert g.atoms[0].Z == 6
    assert g.atoms[1].Z == 6
    assert g.atoms[2].Z == 12
    assert g.atoms.nspecies == 2


def test_re_read(sisl_tmp):
    f = sisl_tmp("file.fdf")
    with open(f, "w") as fh:
        fh.write("Flag1 date\n")
        fh.write("Flag1 not-date\n")
        fh.write("Flag1 not-date-2\n")
        fh.write("Flag3 true\n")

    fdf = fdfSileSiesta(f)
    for i in range(10):
        assert fdf.get("Flag1") == "date"
    assert fdf.get("Flag3")


def test_get_set(sisl_tmp):
    f = sisl_tmp("file.fdf")
    with open(f, "w") as fh:
        fh.write("Flag1 date\n")

    fdf = fdfSileSiesta(f)
    assert fdf.get("Flag1") == "date"
    fdf.set("Flag1", "not-date")
    assert fdf.get("Flag1") == "not-date"
    fdf.set("Flag1", "date")
    assert fdf.get("Flag1") == "date"
    fdf.set("Flag1", "date-date")
    assert fdf.get("Flag1") == "date-date"
    fdf.set("Flag1", "date-date", keep=False)


def test_get_block(sisl_tmp):
    f = sisl_tmp("file.fdf")
    with open(f, "w") as fh:
        fh.write("%block MyBlock\n  date\n%endblock\n")

    fdf = fdfSileSiesta(f)

    assert isinstance(fdf.get("MyBlock"), list)
    assert fdf.get("MyBlock")[0] == "date"
    assert "block" in fdf.print("MyBlock", fdf.get("MyBlock"))


def test_include(sisl_tmp):
    f = sisl_tmp("file.fdf")
    with open(f, "w") as fh:
        fh.write("Flag1 date\n")
        fh.write("# Flag2 comment\n")
        fh.write("Flag2 date2\n")
        fh.write("# Flag3 is read through < from file hello\n")
        fh.write("Flag3 Sub < hello\n")
        fh.write("FakeInt 1\n")
        fh.write("Test 1. eV\n")
        fh.write(" %INCLUDE file2.fdf\n")
        fh.write("TestRy 1. Ry\n")
        fh.write("%block Hello < hello\n")
        fh.write("\n")
        fh.write("TestLast 1. eV\n")

    hello = sisl_tmp("hello")
    with open(hello, "w") as fh:
        fh.write("Flag4 hello\n")
        fh.write("# Comments should be discarded\n")
        fh.write("Flag3 test\n")
        fh.write("Sub sub-test\n")

    file2 = sisl_tmp("file2.fdf")
    with open(file2, "w") as fh:
        fh.write("Flag4 non\n")
        fh.write("\n")
        fh.write("FakeReal 2.\n")
        fh.write("  %incLude file3.fdf")

    file3 = sisl_tmp("file3.fdf")
    with open(file3, "w") as fh:
        fh.write("Sub level\n")
        fh.write("Third level\n")
        fh.write("MyList [1 , 2 , 3]\n")

    fdf = fdfSileSiesta(f, base=sisl_tmp.getbase())
    assert fdf.includes() == [Path(hello), Path(file2), Path(file3)]
    assert fdf.get("Flag1") == "date"
    assert fdf.get("Flag2") == "date2"
    assert fdf.get("Flag3") == "test"
    assert fdf.get("Flag4") == "non"
    assert fdf.get("FLAG4") == "non"
    assert fdf.get("Fakeint") == 1
    assert fdf.get("Fakeint", "0") == "1"
    assert fdf.get("Fakereal") == 2.0
    assert fdf.get("Fakereal", 0.0) == 2.0
    assert fdf.get("test", "eV") == pytest.approx(1.0)
    assert fdf.get("test", with_unit=True)[0] == pytest.approx(1.0)
    assert fdf.get("test", with_unit=True)[1] == "eV"
    assert fdf.get("test", unit="Ry") == pytest.approx(unit_convert("eV", "Ry"))
    assert fdf.get("testRy") == pytest.approx(unit_convert("Ry", "eV"))
    assert fdf.get("testRy", with_unit=True)[0] == pytest.approx(1.0)
    assert fdf.get("testRy", with_unit=True)[1] == "Ry"
    assert fdf.get("testRy", unit="Ry") == pytest.approx(1.0)
    assert fdf.get("Sub") == "sub-test"
    assert fdf.get("Third") == "level"
    assert fdf.get("test-last", with_unit=True)[0] == pytest.approx(1.0)
    assert fdf.get("test-last", with_unit=True)[1] == "eV"

    # Currently lists are not implemented
    # assert np.allclose(fdf.get('MyList'), np.arange(3) + 1)
    # assert np.allclose(fdf.get('MyList', []), np.arange(3) + 1)

    # Read a block
    ll = open(sisl_tmp("hello")).readlines()
    ll.pop(1)
    assert fdf.get("Hello") == [l.replace("\n", "").strip() for l in ll]


def test_xv_preference(sisl_tmp):
    g = geom.graphene()
    g.write(sisl_tmp("file.fdf"))
    g.xyz[0, 0] += 1.0
    g.write(sisl_tmp("siesta.XV"))

    lattice = fdfSileSiesta(sisl_tmp("file.fdf")).read_lattice(True)
    g2 = fdfSileSiesta(sisl_tmp("file.fdf")).read_geometry(True)
    assert np.allclose(lattice.cell, g.cell)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)

    g2 = fdfSileSiesta(sisl_tmp("file.fdf")).read_geometry(order=["fdf"])
    assert np.allclose(g.cell, g2.cell)
    g2.xyz[0, 0] += 1.0
    assert np.allclose(g.xyz, g2.xyz)


def test_geom_order(sisl_tmp):
    pytest.importorskip("netCDF4")
    gfdf = geom.graphene()
    gxv = gfdf.copy()
    gxv.xyz[0, 0] += 0.5
    gnc = gfdf.copy()
    gnc.xyz[0, 0] += 0.5

    gfdf.write(sisl_tmp("siesta.fdf"))

    # Create fdf-file
    fdf = fdfSileSiesta(sisl_tmp("siesta.fdf"))
    assert fdf.read_geometry(order=["nc"]) is None
    gxv.write(sisl_tmp("siesta.XV"))
    gnc.write(sisl_tmp("siesta.nc"))

    # Should read from XV
    g = fdf.read_geometry(True)
    assert np.allclose(g.xyz, gxv.xyz)
    g = fdf.read_geometry(order=["nc", "fdf"])
    assert np.allclose(g.xyz, gnc.xyz)
    g = fdf.read_geometry(order=["fdf", "nc"])
    assert np.allclose(g.xyz, gfdf.xyz)
    g = fdf.read_geometry(True, order="^fdf")
    assert np.allclose(g.xyz, gxv.xyz)


def test_geom_constraints(sisl_tmp):
    gfdf = geom.graphene().tile(2, 0).tile(2, 1)
    gfdf["CONSTRAIN"] = 0
    gfdf["CONSTRAIN-x"] = 2
    gfdf["CONSTRAIN-y"] = [1, 3, 4, 5]
    gfdf["CONSTRAIN-z"] = range(len(gfdf))

    gfdf.write(sisl_tmp("siesta.fdf"))


def test_h2_dynamical_matrix(sisl_files):
    si = fdfSileSiesta(sisl_files("siesta", "H2_hessian", "h2_hessian.fdf"))

    TF = [True, False]

    eV2cm = 8065.54429
    hw_true = [
        -7.11492125e-06,
        1.71779189e-05,
        2.02363231e-05,
        1.25035599e03,
        1.25035599e03,
        3.16727461e03,
    ]

    from itertools import product

    for ti, s0, herm in product(TF, TF, TF):

        dyn = si.read_dynamical_matrix(trans_inv=ti, sum0=s0, hermitian=herm)

        hw = dyn.eigenvalue().hw
        if ti and s0 and herm:
            assert np.allclose(hw * eV2cm, hw_true, atol=1e-4)


def test_dry_read(sisl_tmp):
    # This test runs the read-functions. They aren't expected to actually read anything,
    # it is only a dry-run.
    file = sisl_tmp("siesta.fdf")
    geom.graphene().write(file)
    fdf = fdfSileSiesta(file)

    read_methods = set(m for m in dir(fdf) if m.startswith("read_"))
    output = dict(output=True)
    kwargs = {
        "lattice": output,
        "geometry": output,
        "grid": dict(name="rho"),
    }

    with pytest.warns(SislWarning):
        assert np.allclose(fdf.read_lattice_nsc(), (1, 1, 1))
    read_methods.remove("read_lattice_nsc")

    geom_methods = set(f"read_{x}" for x in ("basis", "lattice", "geometry"))
    read_methods -= geom_methods

    for methodname in read_methods:
        kwarg = kwargs.get(methodname[5:], dict())
        assert getattr(fdf, methodname)(**kwarg) is None

    for methodname in geom_methods:
        # Also run these, but dont assert None due to the graphene values being present
        # in the fdf. The read functions will still go dry-running through eg. nc-files.
        kwarg = kwargs.get(methodname[5:], dict())
        getattr(fdf, methodname)(**kwarg)


def test_fdf_argumentparser(sisl_tmp):
    f = sisl_tmp("file.fdf")
    with open(f, "w") as fh:
        fh.write("Flag1 date\n")
        fh.write("Flag1 not-date\n")
        fh.write("Flag1 not-date-2\n")
        fh.write("Flag3 true\n")

    fdfSileSiesta(f).ArgumentParser()


def test_fdf_fe_basis(sisl_files):
    geom = fdfSileSiesta(
        sisl_files("siesta", "fe_bcc_simple", "fe.fdf")
    ).read_geometry()
    assert geom.na == 1
    # The fe.fdf does not contain basis information, so it will default to 1.
    assert geom.no == 1


def test_fdf_pao_basis():
    fdf = fdfSileSiesta

    block = """
Mg                    1                    # Species label, number of l-shells
 n=3   0   1                         # n, l, Nzeta
   6.620
   1.000
C                     2                    # Species label, number of l-shells
 n=2   0   1                         # n, l, Nzeta
   4.192
   1.000
 n=2   1   1                         # n, l, Nzeta
   4.870
   1.000
O                     2                    # Species label, number of l-shells
 n=2   0   1                         # n, l, Nzeta
   3.305
   1.000
 n=2   1   1                         # n, l, Nzeta
   3.937
   1.000
    """.splitlines()

    atom_orbs = fdf._parse_pao_basis(block)
    assert len(atom_orbs) == 3
    assert len(atom_orbs["Mg"]) == 1
    assert len(atom_orbs["C"]) == 4
    assert len(atom_orbs["O"]) == 4
    for i, (tag, orbs) in enumerate(atom_orbs.items()):
        specie_orbs = fdf._parse_pao_basis(block, species=tag)
        assert specie_orbs == orbs

    block = """
Fe_SOC                2                    # Species label, number of l-shells
 n=4   0   2 P   1                   # n, l, Nzeta, Polarization, NzetaPol
   7.329      6.153
   1.000      1.000
 n=3   2   2                         # n, l, Nzeta
   4.336      2.207
   1.000      1.000
Pt_SOC                2                    # Species label, number of l-shells
 n=6   0   2 P   1                   # n, l, Nzeta, Polarization, NzetaPol
   7.158      6.009
   1.000      1.000
 n=5   2   2                         # n, l, Nzeta
   5.044      3.022
   1.000      1.000
"""
    atom_orbs = fdf._parse_pao_basis(block)
    assert len(atom_orbs) == 2
    assert len(atom_orbs["Fe_SOC"]) == 5 + 10
    assert len(atom_orbs["Pt_SOC"]) == 5 + 10
    for i, (tag, orbs) in enumerate(atom_orbs.items()):
        specie_orbs = fdf._parse_pao_basis(block, species=tag)
        assert specie_orbs == orbs


def test_fdf_gz(sisl_files):
    f = sisl_files("siesta", "fdf", "main.fdf.gz")
    fdf = fdfSileSiesta(f)

    # read from gzipped file
    assert fdf.get("Main.Foo") == "hello"
    assert fdf.get("Main.Bar") == "world"

    # read from included non-gzipped file
    assert fdf.get("Lvl2.Foo") == "world"
    assert fdf.get("Lvl2.Bar") == "hello"

    # read from nested included gzipped file
    assert fdf.get("Lvl3.Foo") == "world3"
    assert fdf.get("Lvl3.Bar") == "hello3"

    f = sisl_files("siesta", "fdf", "level2.fdf")
    fdf = fdfSileSiesta(f)

    # read from non-gzipped file
    assert fdf.get("Lvl2.Foo") == "world"
    assert fdf.get("Lvl2.Bar") == "hello"

    # read from included gzipped file
    assert fdf.get("Lvl3.Foo") == "world3"
    assert fdf.get("Lvl3.Bar") == "hello3"


def test_fdf_block_write_print(sisl_tmp):
    f = sisl_tmp("block_write_print.fdf")
    fdf = fdfSileSiesta(f, "w")

    block_value = ["this is my life"]
    fdf.set("hello", block_value)
    fdf.set("goodbye", "hello")

    fdf = fdfSileSiesta(f)

    assert "hello" == fdf.get("goodbye")
    assert block_value == fdf.get("hello")
    assert f"""%block hello
 {block_value[0]}
%endblock hello
""" == fdf.print(
        "hello", block_value
    )


def test_fdf_write_bandstructure(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.fdf")

    bs = sisl.BandStructure(
        sisl_system.g,
        [[0, 0, 0], [0.5, 0.5, 0.5], [0.25, 0.5, 0]],
        200,
        names=["Gamma", "Edge", "L"],
    )

    with fdfSileSiesta(f, "w") as fdf:
        fdf.write_brillouinzone(bs)

    with fdfSileSiesta(f) as fdf:
        block = fdf.get("BandLines")
    assert len(block) == 3


def test_fdf_read_from_xv(sisl_tmp):
    # test for #778
    f_fdf = sisl_tmp("read_from_xv.fdf")
    sc_lines = [
        "Latticeconstant 1. Ang",
        "%block latticeparameters",
        " 1. 1. 1. 90. 90. 90.",
        "%endblock",
    ]
    lines = [
        "%block chemicalSpeciesLabel",
        " 2 6 C",
        " 1 2 He",
        " 4 3 Li",
        " 3 1 H",  # not present in the geometry
        "%endblock",
        "AtomicCoordinatesFormat Ang",
        "%block atomiccoordinatesandatomicspecies",
        " 1. 1. 1. 2",
        " 0. 0. 1. 1",
        " 1. 0. 1. 4",
        " 1. 1. 1. 2",
        " 1. 0. 1. 4",
        " 0. 0. 1. 1",
        "%endblock",
    ]
    with open(f_fdf, "w") as fh:
        fh.write("\n".join(sc_lines) + "\n")
        fh.write("\n".join(lines))
        fh.write("\nSystemLabel read_from_xv")

    f_xv = sisl_tmp("read_from_xv.XV")
    with open(f_xv, "w") as fh:
        fh.write(
            """\
1. 0. 0.  0. 0. 0.
0. 1. 0.  0. 0. 0.
0. 0. 2.  0. 0. 0.
6
2 6 0. 1. 0.  0. 0. 0.
1 2 0. 1. 1.  0. 0. 0.
4 3 1. 1. 1.  0. 0. 0.
2 6 0. 1. 0.  0. 0. 0.
4 3 1. 1. 1.  0. 0. 0.
1 2 0. 1. 1.  0. 0. 0.
"""
        )

    fdf = fdfSileSiesta(f_fdf, track=True, base=sisl_tmp.getbase())
    geom_fdf = fdf.read_geometry(order="fdf")

    assert len(geom_fdf) == 6
    assert len(geom_fdf.atoms.atom) == 4
    assert np.allclose(geom_fdf.atoms.Z, [6, 2, 3, 6, 3, 2])
    assert np.allclose(geom_fdf.atoms.species, [1, 0, 3, 1, 3, 0])
    assert np.allclose(geom_fdf.xyz[0], [1, 1, 1])
    assert np.allclose(geom_fdf.xyz[1], [0, 0, 1])
    assert np.allclose(geom_fdf.xyz[2], [1, 0, 1])

    geom_xv = fdf.read_geometry(order="xv")
    assert len(geom_xv) == 6
    assert len(geom_xv.atoms.atom) == 4
    assert np.allclose(geom_xv.atoms.Z, [6, 2, 3, 6, 3, 2])
    assert np.allclose(geom_xv.atoms.species, [1, 0, 3, 1, 3, 0])
    xyz = geom_xv.xyz * unit_convert("Ang", "Bohr")
    assert np.allclose(xyz[0], [0, 1, 0])
    assert np.allclose(xyz[1], [0, 1, 1])
    assert np.allclose(xyz[2], [1, 1, 1])


def test_fdf_multiple_atoms_scrambeld(sisl_tmp):
    # test for #778
    f_fdf = sisl_tmp("multiple_atoms_scrambled.fdf")
    sc_lines = [
        "Latticeconstant 1. Ang",
        "%block latticeparameters",
        " 1. 1. 1. 90. 90. 90.",
        "%endblock",
    ]
    lines = [
        "%block chemicalSpeciesLabel",
        " 1 1 H.opt88",
        " 4 6 C.blyp",
        " 2 79 Au.blyp",
        " 3 79 Aus.blyp",
        "%endblock",
        "AtomicCoordinatesFormat Ang",
        "%block atomiccoordinatesandatomicspecies",
        " 1. 1. 1. 1",
        " 0. 0. 1. 2",
        " 1. 0. 1. 3",
        " 1. 1. 1. 4",
        " 1. 0. 1. 1",
        "%endblock",
    ]
    with open(f_fdf, "w") as fh:
        fh.write("\n".join(sc_lines) + "\n")
        fh.write("\n".join(lines))

    geom = fdfSileSiesta(f_fdf).read_geometry()

    assert len(geom) == 5
    assert len(geom.atoms.atom) == 4
    assert np.allclose(geom.atoms.Z, [1, 79, 79, 6, 1])
    assert np.allclose(geom.atoms.species, [0, 1, 2, 3, 0])


def test_fdf_multiple_atoms_linear(sisl_tmp):
    # test for #778
    f = sisl_tmp("multiple_atoms_linear.fdf")
    sc_lines = [
        "Latticeconstant 1. Ang",
        "%block latticeparameters",
        " 1. 1. 1. 90. 90. 90.",
        "%endblock",
    ]
    lines = [
        "%block chemicalSpeciesLabel",
        " 1 1 H.opt88",
        " 2 79 Au.blyp",
        " 3 79 Aus.blyp",
        " 4 6 C.blyp",
        "%endblock",
        "AtomicCoordinatesFormat Ang",
        "%block atomiccoordinatesandatomicspecies",
        " 1. 1. 1. 1",
        " 0. 0. 1. 2",
        " 1. 0. 1. 3",
        " 1. 1. 1. 4",
        " 1. 0. 1. 1",
        "%endblock",
    ]
    with open(f, "w") as fh:
        fh.write("\n".join(sc_lines) + "\n")
        fh.write("\n".join(lines))

    geom = fdfSileSiesta(f).read_geometry()

    assert len(geom) == 5
    assert len(geom.atoms.atom) == 4
    assert np.allclose(geom.atoms.Z, [1, 79, 79, 6, 1])
    assert np.allclose(geom.atoms.species, [0, 1, 2, 3, 0])


def test_fdf_multiple_simple(sisl_tmp):
    # test for #778
    f = sisl_tmp("multiple_simple.fdf")

    with open(f, "w") as fh:
        fh.write(
            """
%block ChemicalSpeciesLabel
    1   79  Au
    2    8   O
%endblock ChemicalSpeciesLabel

LatticeConstant    1.000 Ang
%block LatticeVectors
    4  0  0
    0  10 0
    0  0  10
%endblock LatticeVectors
AtomicCoordinatesFormat Ang
%block AtomicCoordinatesAndAtomicSpecies
    0 0 0  2
    2 0 0  1
%endblock AtomicCoordinatesAndAtomicSpecies
                """
        )

    geom = fdfSileSiesta(f).read_geometry()
    assert geom.na == 2
    assert np.allclose(geom.atoms.Z, [8, 79])
    assert np.allclose(geom.atoms.species, [1, 0])

    with open(f, "w") as fh:
        fh.write(
            """
%block ChemicalSpeciesLabel
    1   79  Au
    2    8   O
%endblock ChemicalSpeciesLabel

LatticeConstant    1.000 Ang
%block LatticeVectors
    4  0  0
    0  10 0
    0  0  10
%endblock LatticeVectors
AtomicCoordinatesFormat Ang
%block AtomicCoordinatesAndAtomicSpecies
    0 0 0  2
    2 0 0  1
    0 0 0  2
%endblock AtomicCoordinatesAndAtomicSpecies
                """
        )

    geom = fdfSileSiesta(f).read_geometry()
    assert geom.na == 3
    assert np.allclose(geom.atoms.Z, [8, 79, 8])
    assert np.allclose(geom.atoms.species, [1, 0, 1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl import Quaternion


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            self.qx = Quaternion(90, [1, 0, 0], rad=False)
            self.qy = Quaternion(90, [0, 1, 0], rad=False)
            self.qz = Quaternion(90, [0, 0, 1], rad=False)
            self.Qx = Quaternion(90, [2, 0, 0], rad=False)
            self.Qy = Quaternion(90, [0, 2, 0], rad=False)
            self.Qz = Quaternion(90, [0, 0, 2], rad=False)

    return t()


@pytest.mark.quaternion
class TestQuaternion:
    def test_copy(self, setup):
        qx = setup.qx.copy()
        assert qx == setup.qx
        assert qx._v is not setup.qx._v.base

    def test_init(self):
        a = Quaternion(90, [1, 0, 1])
        b = Quaternion([1, 0, 1], 90)
        assert a == b
        c = Quaternion([1, 0, 2, 1])
        assert a != c

    def test_str_repr(self):
        a = Quaternion(90, [1, 0, 1])
        assert str(a) != repr(a)

    def test_conj(self, setup):
        qx = setup.qx.conj()
        assert qx.conj() == setup.qx

    def test_norm(self, setup):
        for c in "xyz":
            assert getattr(setup, "q" + c).norm() == pytest.approx(1.0)

    def test_degree1(self, setup):
        for c in "xyz":
            assert getattr(setup, "q" + c).angle(in_rad=False) == pytest.approx(90)

    def test_radians1(self, setup):
        rx = setup.qx.radian
        ry = setup.qy.radian
        assert rx == ry

    def test_op1(self, setup):
        rx = -setup.qx
        assert -rx == setup.qx

        rxy = setup.qx + setup.qy
        setup.qx += setup.qy
        assert rxy == setup.qx
        setup.qx -= setup.qy
        assert -rx == setup.qx

        rxy = setup.qx - setup.qy
        setup.qx -= setup.qy
        assert rxy == setup.qx
        setup.qx += setup.qy
        assert -rx == setup.qx

    def test_op2(self, setup):
        rx = setup.qx + 1.0
        assert rx - 1.0 == setup.qx

        rx = setup.qx * 1.0
        assert rx == setup.qx

        rx = setup.qx * 1.0
        assert rx == setup.qx

        rx = setup.qx / 1.0
        assert rx == setup.qx

        rx = setup.qx.copy()
        rx += 1.0
        rx -= 1.0
        assert rx == setup.qx

        rx = setup.qx.copy()
        rx *= setup.qy
        assert rx == setup.qx * setup.qy

    def test_fail_div1(self, setup):
        with pytest.raises(NotImplementedError):
            setup.qx /= setup.qy

    def test_fail_div2(self, setup):
        with pytest.raises(NotImplementedError):
            a = setup.qx / setup.qy



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from sisl.viz.processors.xarray import group_reduce

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(scope="module")
def dataarray():
    return xr.DataArray(
        [
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
            [10, 11, 12],
        ],
        coords=[("x", [0, 1, 2, 3]), ("y", [0, 1, 2])],
        name="vals",
    )


@pytest.fixture(scope="module")
def dataset():
    arr = xr.DataArray(
        [
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
            [10, 11, 12],
        ],
        coords=[("x", [0, 1, 2, 3]), ("y", [0, 1, 2])],
        name="vals",
    )

    arr2 = arr * 2
    return xr.Dataset({"vals": arr, "double": arr2})


def test_dataarray(dataarray):
    new = group_reduce(
        dataarray,
        [{"selector": [0, 1]}, {"selector": [2, 3]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]
    assert new.sel(selection=0).sum() == 1 + 2 + 3 + 4 + 5 + 6
    assert new.sel(selection=1).sum() == 7 + 8 + 9 + 10 + 11 + 12


def test_dataarray_multidim(dataarray):
    new = group_reduce(
        dataarray,
        [{"selector": ([0, 1], [0, 1])}, {"selector": ([2, 3], [0, 1])}],
        reduce_dim=("x", "y"),
        reduce_func=np.sum,
        groups_dim="selection",
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "y" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]
    assert new.sel(selection=0).sum() == 1 + 2 + 4 + 5
    assert new.sel(selection=1).sum() == 7 + 8 + 10 + 11


def test_dataarray_multidim_multireduce(dataarray):
    new = group_reduce(
        dataarray,
        [{"selector": ([0, 1], [0, 1])}, {"selector": ([2, 3], [0, 1])}],
        reduce_dim=("x", "y"),
        reduce_func=(np.sum, np.mean),
        groups_dim="selection",
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "y" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]
    assert new.sel(selection=0).sum() == (1 + 4) / 2 + (2 + 5) / 2
    assert new.sel(selection=1).sum() == (7 + 10) / 2 + (8 + 11) / 2


def test_dataarray_sangroup(dataarray):
    # We use sanitize group to simply set all selectors to [0,1]
    new = group_reduce(
        dataarray,
        [{"selector": [0, 1]}, {"selector": [2, 3]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        sanitize_group=lambda group: {**group, "selector": [0, 1]},
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]
    assert new.sel(selection=0).sum() == 1 + 2 + 3 + 4 + 5 + 6
    assert new.sel(selection=1).sum() == 1 + 2 + 3 + 4 + 5 + 6


def test_dataarray_names(dataarray):
    new = group_reduce(
        dataarray,
        [{"selector": [0, 1], "name": "first"}, {"selector": [2, 3], "name": "second"}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == ["first", "second"]
    assert new.sel(selection="first").sum() == 1 + 2 + 3 + 4 + 5 + 6
    assert new.sel(selection="second").sum() == 7 + 8 + 9 + 10 + 11 + 12


def test_dataarray_groupvars(dataarray):
    new = group_reduce(
        dataarray,
        [
            {"selector": [0, 1], "name": "first", "color": "red", "size": 3},
            {"selector": [2, 3], "name": "second", "color": "blue", "size": 4},
        ],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        group_vars=["color", "size"],
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == ["first", "second"]

    assert "vals" in new
    assert new.vals.sel(selection="first").sum() == 1 + 2 + 3 + 4 + 5 + 6
    assert new.vals.sel(selection="second").sum() == 7 + 8 + 9 + 10 + 11 + 12

    for k, vals in {"color": ["red", "blue"], "size": [3, 4]}.items():
        assert k in new

        k_data = getattr(new, k)
        assert "selection" in k_data.dims
        assert list(k_data.coords["selection"]) == ["first", "second"]
        assert list(k_data) == vals


@pytest.mark.parametrize("drop", [True, False])
def test_dataarray_empty_selector(dataarray, drop):
    new = group_reduce(
        dataarray,
        [{"selector": []}, {"selector": [2, 3]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        drop_empty=drop,
        fill_empty=0.0,
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "selection" in new.dims
    if drop:
        assert len(new.coords["selection"]) == 1
        assert list(new.coords["selection"]) == [1]
    else:
        assert len(new.coords["selection"]) == 2
        assert list(new.coords["selection"]) == [0, 1]

    if not drop:
        assert new.sel(selection=0).sum() == 0.0
    assert new.sel(selection=1).sum() == 7 + 8 + 9 + 10 + 11 + 12


def test_dataarray_empty_selector_0d():
    """When reducing an array along its only dimension, you get a 0d array.

    This was creating an error when filling empty selections. This test
    ensures that it doesn' happen again
    """

    new = group_reduce(
        xr.DataArray([1, 2, 3], coords=[("x", [0, 1, 2])]),
        [{"selector": []}, {"selector": [1, 2]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        drop_empty=False,
        fill_empty=0.0,
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]

    assert new.sel(selection=0).sum() == 0.0
    assert new.sel(selection=1).sum() == 2 + 3


def test_dataset(dataset):
    new = group_reduce(
        dataset,
        [{"selector": [0, 1]}, {"selector": [2, 3]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]

    assert "vals" in new
    assert "double" in new
    assert new.sel(selection=0).sum() == (1 + 2 + 3 + 4 + 5 + 6) * 3
    assert new.sel(selection=1).sum() == (7 + 8 + 9 + 10 + 11 + 12) * 3


def test_dataset_multidim(dataset):
    new = group_reduce(
        dataset,
        [{"selector": ([0, 1], [0, 1])}, {"selector": ([2, 3], [0, 1])}],
        reduce_dim=("x", "y"),
        reduce_func=np.sum,
        groups_dim="selection",
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "y" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]

    assert "vals" in new
    assert "double" in new
    assert new.sel(selection=0).sum() == (1 + 2 + 4 + 5) * 3
    assert new.sel(selection=1).sum() == (7 + 8 + 10 + 11) * 3


def test_dataset_multidim_multireduce(dataset):
    new = group_reduce(
        dataset,
        [{"selector": ([0, 1], [0, 1])}, {"selector": ([2, 3], [0, 1])}],
        reduce_dim=("x", "y"),
        reduce_func=(np.sum, np.mean),
        groups_dim="selection",
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "y" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]

    assert "vals" in new
    assert "double" in new
    assert new.sel(selection=0).sum() == ((1 + 4) / 2 + (2 + 5) / 2) * 3
    assert new.sel(selection=1).sum() == ((7 + 10) / 2 + (8 + 11) / 2) * 3


def test_dataarray_multidim_multireduce(dataarray):
    new = group_reduce(
        dataarray,
        [{"selector": ([0, 1], [0, 1])}, {"selector": ([2, 3], [0, 1])}],
        reduce_dim=("x", "y"),
        reduce_func=(np.sum, np.mean),
        groups_dim="selection",
    )

    assert isinstance(new, xr.DataArray)
    assert "x" not in new.dims
    assert "y" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]
    assert new.sel(selection=0).sum() == (1 + 4) / 2 + (2 + 5) / 2
    assert new.sel(selection=1).sum() == (7 + 10) / 2 + (8 + 11) / 2


def test_dataset_names(dataset):
    new = group_reduce(
        dataset,
        [{"selector": [0, 1], "name": "first"}, {"selector": [2, 3], "name": "second"}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == ["first", "second"]

    assert "vals" in new
    assert "double" in new
    assert new.sel(selection="first").sum() == (1 + 2 + 3 + 4 + 5 + 6) * 3
    assert new.sel(selection="second").sum() == (7 + 8 + 9 + 10 + 11 + 12) * 3


def test_dataset_groupvars(dataset):
    new = group_reduce(
        dataset,
        [
            {"selector": [0, 1], "name": "first", "color": "red", "size": 3},
            {"selector": [2, 3], "name": "second", "color": "blue", "size": 4},
        ],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        group_vars=["color", "size"],
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == ["first", "second"]

    assert "vals" in new
    assert "double" in new
    assert new.double.sel(selection="first").sum() == (1 + 2 + 3 + 4 + 5 + 6) * 2
    assert new.double.sel(selection="second").sum() == (7 + 8 + 9 + 10 + 11 + 12) * 2

    for k, vals in {"color": ["red", "blue"], "size": [3, 4]}.items():
        assert k in new

        k_data = getattr(new, k)
        assert "selection" in k_data.dims
        assert list(k_data.coords["selection"]) == ["first", "second"]
        assert list(k_data) == vals


@pytest.mark.parametrize("drop", [True, False])
def test_dataset_empty_selector(dataset, drop):
    new = group_reduce(
        dataset,
        [{"selector": []}, {"selector": [2, 3]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        drop_empty=drop,
        fill_empty=0.0,
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "selection" in new.dims
    if drop:
        assert len(new.coords["selection"]) == 1
        assert list(new.coords["selection"]) == [1]
    else:
        assert len(new.coords["selection"]) == 2
        assert list(new.coords["selection"]) == [0, 1]

    assert "vals" in new
    assert "double" in new
    if not drop:
        assert new.sel(selection=0).sum() == 0.0
    assert new.sel(selection=1).sum() == (7 + 8 + 9 + 10 + 11 + 12) * 3


def test_dataaset_empty_selector_0d():
    """When reducing an array along its only dimension, you get a 0d array.

    This was creating an error when filling empty selections. This test
    ensures that it doesn' happen again
    """
    dataset = xr.Dataset({"vals": (["x"], [1, 2, 3])})

    new = group_reduce(
        dataset,
        [{"selector": []}, {"selector": [1, 2]}],
        reduce_dim="x",
        reduce_func=np.sum,
        groups_dim="selection",
        drop_empty=False,
        fill_empty=0.0,
    )

    assert isinstance(new, xr.Dataset)
    assert "x" not in new.dims
    assert "selection" in new.dims
    assert len(new.coords["selection"]) == 2
    assert list(new.coords["selection"]) == [0, 1]

    assert "vals" in new
    assert new.sel(selection=0).sum() == 0.0
    assert new.sel(selection=1).sum() == 2 + 3



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_fe(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "fe_bcc_simple", "fe.bands"))
    labels, k, eig = si.read_data()
    assert k.shape == (131,)
    assert eig.shape == (131, 2, 19)
    assert eig[0, 0, 0] == pytest.approx(-93.9944 + 5.9979689568167363)
    assert len(labels[0]) == 5


def test_fe_ArgumentParser(sisl_files, sisl_tmp):
    pytest.importorskip("matplotlib", reason="matplotlib not available")
    png = sisl_tmp("fe.bands.png")
    si = sisl.get_sile(sisl_files("siesta", "fe_bcc_simple", "fe.bands"))
    p, ns = si.ArgumentParser()
    p.parse_args([], namespace=ns)
    p.parse_args(["--energy", " -2:2"], namespace=ns)
    p.parse_args(["--energy", " -2:2", "--plot", png], namespace=ns)


def test_fe_xarray(sisl_files, sisl_tmp):
    pytest.importorskip("xarray", reason="xarray not available")
    si = sisl.get_sile(sisl_files("siesta", "fe_bcc_simple", "fe.bands"))

    bands = si.read_data(as_dataarray=True)
    assert len(bands["k"]) == 131
    assert len(bands["spin"]) == 2
    assert len(bands["band"]) == 19
    assert len(bands.ticks) == len(bands.ticklabels) == 5



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import scipy.linalg as sl

from sisl.linalg import solve, solve_destroy

pytestmark = [pytest.mark.linalg, pytest.mark.eig]


def test_solve1():
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    b = np.random.rand(10, 10)
    xs = sl.solve(a, b)
    x = solve(a, b)
    assert np.allclose(xs, x)


def test_solve2():
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    ac = a.copy()
    b = np.random.rand(10)
    bc = b.copy()
    xs = sl.solve(a, b)
    x = solve(a, b)
    assert np.allclose(xs, x)
    assert x.shape == (10,)
    assert np.allclose(a, ac)
    assert np.allclose(b, bc)


def test_solve3():
    np.random.seed(1285947159)
    a = np.random.rand(10, 2)
    b = np.random.rand(10)
    with pytest.raises(ValueError):
        solve(a, b)


def test_solve4():
    np.random.seed(1285947159)
    a = np.random.rand(10, 10)
    b = np.random.rand(10)
    xs = sl.solve(a, b)
    x = solve_destroy(a, b)
    assert np.allclose(xs, x)
    assert x.shape == (10,)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl
from sisl import Grid
from sisl.viz.plots import grid_plot

pytestmark = [pytest.mark.viz, pytest.mark.plots]


@pytest.fixture(scope="module", params=["plotly", "matplotlib"])
def backend(request):
    pytest.importorskip(request.param)
    return request.param


@pytest.fixture(scope="module", params=["x", "xy", "xyz"])
def axes(request):
    return request.param


@pytest.fixture(scope="module")
def grid():
    geometry = sisl.geom.graphene()
    grid = Grid((10, 10, 10), geometry=geometry)

    grid.grid[:] = np.linspace(0, 1000, 1000).reshape(10, 10, 10)
    return grid


def test_grid_plot(grid, axes, backend):
    if axes == "xyz" and backend == "matplotlib":
        with pytest.raises(NotImplementedError):
            grid_plot(grid, axes=axes, backend=backend)
    else:
        grid_plot(grid, axes=axes, backend=backend)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.utils.mathematics import *

pytestmark = pytest.mark.utils


def test_curl_2d():
    a = np.random.rand(3, 3)
    C = curl(a)
    assert C.shape == (3,)
    c = [a[1, 2] - a[2, 1], a[2, 0] - a[0, 2], a[0, 1] - a[1, 0]]
    assert np.allclose(C, c)


def test_curl_3d():
    a = np.random.rand(4, 3, 3)
    C = curl(a)
    assert C.shape == (4, 3)


def test_curl_4d():
    a = np.random.rand(4, 3, 4, 3)
    C1 = curl(a, axis=1, axisv=3)
    C2 = curl(a, axis=1)
    assert C1.shape == (4, 4, 3)
    assert np.allclose(C1, C2)

    b = np.swapaxes(a, 1, 2)
    C3 = curl(b)
    # no need to swap, we are removing the 2nd last axis
    assert np.allclose(C1, C3)


def test_curl_6d():
    a = np.random.rand(4, 3, 4, 10, 3, 3)
    C1 = curl(a, axis=1, axisv=5)
    assert C1.shape == (4, 4, 10, 3, 3)
    C2 = curl(a, axis=1)
    assert C2.shape == (4, 4, 10, 3, 3)
    assert np.allclose(C1, C2)

    C2 = curl(a, axis=-1, axisv=1)
    assert C2.shape == (4, 3, 4, 10, 3)


def test_curl_4d_same():
    # same axis specification
    a = np.random.rand(4, 3, 4, 3)
    with pytest.raises(ValueError):
        curl(a, axis=1, axisv=1)


def test_curl_4d_not_3():
    # axis lengths must equal 3 (nothing else is supported currently)
    a = np.random.rand(4, 3, 4, 3)
    with pytest.raises(ValueError):
        curl(a, axis=1, axisv=2)


@pytest.mark.parametrize("nd", [0, 1, 2, 3])
def test_cart2spher_nd(nd):
    r = np.random.rand(*([3] * (nd + 1)))
    R, theta, phi = cart2spher(r)
    assert R.ndim == nd
    assert theta.ndim == nd
    assert phi.ndim == nd


def test_cart2spher_nd_maxr():
    r = np.random.rand(3, 3)
    idx, R, theta, phi = cart2spher(r, maxR=0.5)
    assert idx.ndim == 1
    assert R.ndim == 1
    assert theta.ndim == 1
    assert phi.ndim == 1


def test_rotation_matrix():
    angles = [10, 40, 59]
    R1 = rotation_matrix(*angles, order="xyz")
    R2 = rotation_matrix(*list(map(lambda x: -x, angles)), order="zyx")
    assert np.allclose(R1 @ R2, np.identity(3))

    R1 = rotation_matrix(*angles, order="xz")
    R2 = rotation_matrix(*list(map(lambda x: -x, angles)), order="zx")
    assert np.allclose(R1 @ R2, np.identity(3))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl
from sisl.io.vasp.outcar import outcarSileVASP

pytestmark = [pytest.mark.io, pytest.mark.vasp]


def test_diamond_outcar_energies(sisl_files):
    f = sisl_files("vasp", "diamond", "OUTCAR")
    f = outcarSileVASP(f)

    with pytest.warns(sisl.SislDeprecation, match=r"no longer returns the last entry"):
        E0 = f.read_energy()

        E = f.read_energy[-1]()
        Eall = f.read_energy[:]()

    assert E0.sigma0 == 38.32164289  # first block
    assert E.sigma0 == -18.16472605  # last block

    assert E0 == Eall[0]
    assert E == Eall[-1]
    assert len(Eall) > 1
    assert f.info.completed()

    EHa = f.read_energy[-1](units="Ha")
    assert EHa.sigma0 != E.sigma0


def test_diamond_outcar_cputime(sisl_files):
    f = sisl_files("vasp", "diamond", "OUTCAR")
    f = outcarSileVASP(f)

    assert f.cpu_time() > 0.0
    assert f.info.completed()


def test_diamond_outcar_completed(sisl_files):
    f = sisl_files("vasp", "diamond", "OUTCAR")
    f = outcarSileVASP(f)

    assert f.info.completed()


def test_diamond_outcar_trajectory(sisl_files):
    f = sisl_files("vasp", "diamond", "OUTCAR")
    f = outcarSileVASP(f)

    step = f.read_trajectory()

    assert step.xyz[0, 1] == 0.0
    assert step.xyz[1, 1] == 0.89250
    assert step.force[0, 1] == 0.0
    assert step.force[1, 0] == 0.0

    traj = f.read_trajectory[:]()
    assert len(traj) == 1


def test_graphene_relax_outcar_trajectory(sisl_files):
    f = sisl_files("vasp", "graphene_relax", "OUTCAR")
    f = outcarSileVASP(f)

    step = f.read_trajectory[9]()
    assert step.cell[0, 0] == 2.472466059
    assert step.cell[1, 0] == -1.236233029
    assert step.cell[2, 2] == 9.789010603
    assert step.force[0, 2] == -0.03229
    assert step.force[1, 2] == 0.03229

    traj = f.read_trajectory[:]()
    assert len(traj) == 11
    assert traj[0].cell[0, 0] == 2.441046239
    assert traj[0].xyz[0, 2] == 0.00000
    assert traj[0].xyz[1, 2] == 0.20000
    assert traj[0].force[0, 2] == 3.353225
    assert traj[0].force[1, 2] == -3.353225
    assert traj[-1].cell[2, 2] == 9.791779146
    assert traj[-1].xyz[1, 2] == 9.79061
    assert traj[-1].force[0, 2] == -0.019886
    assert traj[-1].force[1, 2] == 0.019886


def test_graphene_md_outcar_trajectory(sisl_files):
    f = sisl_files("vasp", "graphene_md", "OUTCAR")
    f = outcarSileVASP(f)

    step = f.read_trajectory[99]()
    assert step.xyz[0, 0] == 0.09669
    assert step.force[0, 2] == -0.196454

    traj = f.read_trajectory[:]()
    assert len(traj) == 100
    assert traj[0].xyz[0, 0] == 0.12205
    assert traj[0].xyz[1, 0] == 0.09520
    assert traj[0].force[0, 1] == -0.022846
    assert traj[0].force[1, 1] == 0.022846
    assert traj[-1].xyz[0, 0] == 0.09669
    assert traj[-1].force[0, 2] == -0.196454
